from __future__ import annotations

import operator as ops

import pytest

from daft.datatype import DataType
from daft.expressions import col
from tests.expressions.typing.conftest import (
    assert_typing_resolve_vs_runtime_behavior,
    is_integer,
    is_signed_integer,
)


def logical_resolvable(lhs: DataType, rhs: DataType) -> bool:
    # Must have a Bool on one side; may have a Bool or Null on the other.
    if is_integer(lhs) and is_integer(rhs):
        if (lhs == DataType.uint64() and is_signed_integer(rhs)) or (
            rhs == DataType.uint64() and is_signed_integer(lhs)
        ):
            return False
        return True
    elif (is_integer(lhs) and rhs == DataType.null()) or (is_integer(rhs) and lhs == DataType.null()):
        return True
    else:
        return {lhs, rhs} in ({DataType.bool()}, {DataType.bool(), DataType.null()})


@pytest.mark.parametrize("op", [ops.and_, ops.or_])
def test_logical_binary(binary_data_fixture, op):
    lhs, rhs = binary_data_fixture
    assert_typing_resolve_vs_runtime_behavior(
        data=binary_data_fixture,
        expr=op(col(lhs.name()), col(rhs.name())),
        run_kernel=lambda: op(lhs, rhs),
        resolvable=logical_resolvable(lhs.datatype(), rhs.datatype()),
    )


@pytest.mark.parametrize("op", [ops.invert])
def test_logical_unary(unary_data_fixture, op):
    arg = unary_data_fixture
    assert_typing_resolve_vs_runtime_behavior(
        data=(unary_data_fixture,),
        expr=op(col(arg.name())),
        run_kernel=lambda: op(arg),
        resolvable=arg.datatype() == DataType.bool(),
    )



from __future__ import annotations

import numpy as np
import pyarrow as pa
import pytest

import daft
from tests.conftest import get_tests_daft_runner_name


class MockException(Exception):
    pass


@pytest.mark.parametrize("materialized", [False, True])
def test_iter_rows(make_df, materialized):
    # Test that df.__iter__ produces the correct rows in the correct order.
    # It should work regardless of whether the dataframe has already been materialized or not.

    df = make_df({"a": list(range(10))}).into_partitions(5).with_column("b", daft.col("a") + 100)
    if materialized:
        df = df.collect()

    rows = list(iter(df))
    assert rows == [{"a": x, "b": x + 100} for x in range(10)]


@pytest.mark.parametrize(
    "format, data, expected",
    [
        ### Ints
        pytest.param("python", [1, 2, 3], [1, 2, 3], id="python_ints"),
        pytest.param(
            "arrow",
            [1, 2, 3],
            [pa.scalar(1), pa.scalar(2), pa.scalar(3)],
            id="arrow_ints",
        ),
        ### Strings
        pytest.param("python", ["a", "b", "c"], ["a", "b", "c"], id="python_strs"),
        pytest.param(
            "arrow",
            ["a", "b", "c"],
            [
                pa.scalar("a", pa.large_string()),
                pa.scalar("b", pa.large_string()),
                pa.scalar("c", pa.large_string()),
            ],
            id="arrow_strs",
        ),
        ### Lists
        pytest.param("python", [[1, 2], [3, 4]], [[1, 2], [3, 4]], id="python_lists"),
        pytest.param(
            "arrow",
            [[1, 2], [3, 4]],
            [
                pa.scalar([1, 2], pa.large_list(pa.int64())),
                pa.scalar([3, 4], pa.large_list(pa.int64())),
            ],
            id="arrow_lists",
        ),
        ### Structs
        pytest.param(
            "python",
            [{"a": 1, "b": 2}, {"a": 3, "b": 4}],
            [{"a": 1, "b": 2}, {"a": 3, "b": 4}],
            id="python_structs",
        ),
        pytest.param(
            "arrow",
            [{"a": 1, "b": 2}, {"a": 3, "b": 4}],
            [
                pa.scalar(
                    {"a": 1, "b": 2},
                    pa.struct([pa.field("a", pa.int64()), pa.field("b", pa.int64())]),
                ),
                pa.scalar(
                    {"a": 3, "b": 4},
                    pa.struct([pa.field("a", pa.int64()), pa.field("b", pa.int64())]),
                ),
            ],
            id="arrow_structs",
        ),
    ],
)
def test_iter_rows_column_formats(make_df, format, data, expected):
    df = make_df({"a": data})

    rows = list(df.iter_rows(column_format=format))

    # Compare each row
    assert len(rows) == len(expected)
    for actual_row, expected_row in zip(rows, [{"a": e} for e in expected]):
        assert actual_row == expected_row


@pytest.mark.parametrize(
    "data, expected",
    [
        pytest.param(
            [[1, 2], [3, 4]],
            [
                np.array([1, 2], dtype=np.int64),
                np.array([3, 4], dtype=np.int64),
            ],
            id="list_of_ints",
        ),
        pytest.param(
            [[1.0, 2.0], [3.0, 4.0]],
            [
                np.array([1.0, 2.0], dtype=np.float64),
                np.array([3.0, 4.0], dtype=np.float64),
            ],
            id="list_of_floats",
        ),
    ],
)
def test_iter_rows_lists_to_numpy(make_df, data, expected):
    df = make_df({"a": data})

    rows = list(df.iter_rows(column_format="arrow"))

    # Compare each row
    assert len(rows) == len(expected)
    for actual_row, expected_row in zip(rows, [{"a": e} for e in expected]):
        np_data = actual_row["a"].values.to_numpy()
        np.testing.assert_array_equal(np_data, expected_row["a"])


def test_iter_rows_arrow_column_format_not_compatible():
    df = daft.from_pydict({"a": [object()]})  # Object type is not supported by arrow or numpy

    with pytest.raises(ValueError):
        list(df.iter_rows(column_format="arrow"))


@pytest.mark.parametrize("materialized", [False, True])
def test_iter_partitions(make_df, materialized):
    # Test that df.iter_partitions() produces partitions in the correct order.
    # It should work regardless of whether the dataframe has already been materialized or not.

    with daft.execution_config_ctx(default_morsel_size=2):
        df = make_df({"a": list(range(10))}).into_partitions(5).with_column("b", daft.col("a") + 100)

        if materialized:
            df = df.collect()

        parts = list(df.iter_partitions())
        if get_tests_daft_runner_name() == "ray":
            import ray

            parts = ray.get(parts)
        parts = [_.to_pydict() for _ in parts]

        assert parts == [
            {"a": [0, 1], "b": [100, 101]},
            {"a": [2, 3], "b": [102, 103]},
            {"a": [4, 5], "b": [104, 105]},
            {"a": [6, 7], "b": [106, 107]},
            {"a": [8, 9], "b": [108, 109]},
        ]


def test_iter_exception(make_df):
    # Test that df.__iter__ actually returns results before completing execution.
    # We test this by raising an exception in a UDF if too many partitions are executed.

    @daft.udf(return_dtype=daft.DataType.int64())
    def echo_or_trigger(s):
        trigger = max(s.to_pylist())
        if trigger >= 199:
            raise MockException(trigger)
        else:
            return s

    with daft.execution_config_ctx(default_morsel_size=2):
        df = make_df({"a": list(range(200))}).into_partitions(100).with_column("b", echo_or_trigger(daft.col("a")))

        it = iter(df)
        assert next(it) == {"a": 0, "b": 0}

        # Ensure the exception does trigger if execution continues.
        with pytest.raises(MockException) as exc_info:
            list(it)

        # Ray's wrapping of the exception loses information about the `.cause`, but preserves it in the string error message
        if get_tests_daft_runner_name() == "ray":
            assert "MockException" in str(exc_info.value)
        assert str(exc_info.value).endswith("failed when executing on inputs:\n  - a (Int64, length=2)")


def test_iter_partitions_exception(make_df):
    # Test that df.iter_partitions actually returns results before completing execution.
    # We test this by raising an exception in a UDF if too many partitions are executed.

    @daft.udf(return_dtype=daft.DataType.int64())
    def echo_or_trigger(s):
        trigger = max(s.to_pylist())
        if trigger >= 199:
            raise MockException(trigger)
        else:
            return s

    with daft.execution_config_ctx(default_morsel_size=2):
        df = make_df({"a": list(range(200))}).into_partitions(100).with_column("b", echo_or_trigger(daft.col("a")))

        it = df.iter_partitions()
        part = next(it)
        if get_tests_daft_runner_name() == "ray":
            import ray

            part = ray.get(part)
        part = part.to_pydict()

        assert part == {"a": [0, 1], "b": [0, 1]}

        # Ensure the exception does trigger if execution continues.
        with pytest.raises(MockException) as exc_info:
            res = list(it)
            if get_tests_daft_runner_name() == "ray":
                ray.get(res)

        # Ray's wrapping of the exception loses information about the `.cause`, but preserves it in the string error message
        if get_tests_daft_runner_name() == "ray":
            assert "MockException" in str(exc_info.value)
        assert str(exc_info.value).endswith("failed when executing on inputs:\n  - a (Int64, length=2)")



from __future__ import annotations

from collections.abc import Iterator
from typing import TYPE_CHECKING, Callable

from daft.daft import CsvParseOptions, JsonParseOptions, PySchema
from daft.daft import PyField as _PyField
from daft.daft import PySchema as _PySchema
from daft.daft import read_csv_schema as _read_csv_schema
from daft.daft import read_json_schema as _read_json_schema
from daft.daft import read_parquet_schema as _read_parquet_schema
from daft.datatype import DataType, TimeUnit, _ensure_registered_super_ext_type

if TYPE_CHECKING:
    from collections.abc import Iterator

    import pyarrow as pa

    from daft.io import IOConfig


__all__ = [
    "DataType",
    "Field",
    "Schema",
]


class Field:
    _field: _PyField

    def __init__(self) -> None:
        raise NotImplementedError("We do not support creating a Field via __init__ ")

    @staticmethod
    def _from_pyfield(field: _PyField) -> Field:
        f = Field.__new__(Field)
        f._field = field
        return f

    @staticmethod
    def create(name: str, dtype: DataType) -> Field:
        pyfield = _PyField.create(name, dtype._dtype)
        return Field._from_pyfield(pyfield)

    @property
    def name(self) -> str:
        return self._field.name()

    @property
    def dtype(self) -> DataType:
        return DataType._from_pydatatype(self._field.dtype())

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Field):
            return False
        return self._field.eq(other._field)

    def __repr__(self) -> str:
        return f"Field(name={self.name}, dtype={self.dtype})"


class Schema:
    _schema: _PySchema

    def __init__(self) -> None:
        raise NotImplementedError("We do not support creating a Schema via __init__ ")

    @staticmethod
    def _from_pyschema(schema: _PySchema) -> Schema:
        s = Schema.__new__(Schema)
        s._schema = schema
        return s

    @classmethod
    def from_pyarrow_schema(cls, pa_schema: pa.Schema) -> Schema:
        """Creates a Daft Schema from a PyArrow Schema.

        Args:
            pa_schema (pa.Schema): PyArrow schema to convert

        Returns:
            Schema: Converted Daft schema
        """
        return cls._from_field_name_and_types(
            [(pa_field.name, DataType.from_arrow_type(pa_field.type)) for pa_field in pa_schema]
        )

    def to_pyarrow_schema(self) -> pa.Schema:
        """Converts a Daft Schema to a PyArrow Schema.

        Returns:
            pa.Schema: PyArrow schema that corresponds to the provided Daft schema
        """
        _ensure_registered_super_ext_type()
        return self._schema.to_pyarrow_schema()

    @classmethod
    def _from_field_name_and_types(self, fields: list[tuple[str, DataType]]) -> Schema:
        assert isinstance(fields, list), f"Expected a list of field tuples, received: {fields}"
        for field in fields:
            assert isinstance(field, tuple), f"Expected a tuple, received: {field}"
            assert len(field) == 2, f"Expected a tuple of length 2, received: {field}"
            name, dtype = field
            assert isinstance(name, str), f"Expected a string name, received: {name}"
            assert isinstance(dtype, DataType), f"Expected a DataType dtype, received: {dtype}"

        s = Schema.__new__(Schema)
        s._schema = _PySchema.from_field_name_and_types([(name, dtype._dtype) for name, dtype in fields])
        return s

    @classmethod
    def _from_fields(self, fields: list[Field]) -> Schema:
        s = Schema.__new__(Schema)
        s._schema = _PySchema.from_fields([f._field for f in fields])
        return s

    def __getitem__(self, key: str) -> Field:
        assert isinstance(key, str), f"Expected str for key, but received: {type(key)}"
        if key not in self._schema.names():
            raise ValueError(f"{key} was not found in Schema of fields {[f.name for f in self]}")
        pyfield = self._schema[key]
        return Field._from_pyfield(pyfield)

    def __len__(self) -> int:
        return len(self._schema.names())

    def column_names(self) -> list[str]:
        return list(self._schema.names())

    def estimate_row_size_bytes(self) -> float:
        return self._schema.estimate_row_size_bytes()

    def __iter__(self) -> Iterator[Field]:
        col_names = self.column_names()
        yield from (self[name] for name in col_names)

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Schema) and self._schema == other._schema

    def to_name_set(self) -> set[str]:
        return set(self.column_names())

    def __repr__(self) -> str:
        return repr(self._schema)

    def _repr_html_(self) -> str:
        return self._schema._repr_html_()

    def _truncated_table_html(self) -> str:
        return self._schema._truncated_table_html()

    def _truncated_table_string(self) -> str:
        return self._schema._truncated_table_string()

    def apply_hints(self, hints: Schema) -> Schema:
        return Schema._from_pyschema(self._schema.apply_hints(hints._schema))

    # Takes the unions between two schemas. Throws an error if the schemas contain overlapping keys.
    def union(self, other: Schema) -> Schema:
        if not isinstance(other, Schema):
            raise ValueError(f"Expected Schema, got other: {type(other)}")

        return Schema._from_pyschema(self._schema.union(other._schema))

    def __reduce__(self) -> tuple[Callable[[PySchema], Schema], tuple[PySchema]]:
        return Schema._from_pyschema, (self._schema,)

    @classmethod
    def from_pydict(cls, fields: dict[str, DataType]) -> Schema:
        return cls._from_fields([Field.create(k, v) for k, v in fields.items()])

    @classmethod
    def from_parquet(
        cls,
        path: str,
        io_config: IOConfig | None = None,
        multithreaded_io: bool | None = None,
        coerce_int96_timestamp_unit: TimeUnit = TimeUnit.ns(),
    ) -> Schema:
        return Schema._from_pyschema(
            _read_parquet_schema(
                uri=path,
                io_config=io_config,
                multithreaded_io=multithreaded_io,
                coerce_int96_timestamp_unit=coerce_int96_timestamp_unit._timeunit,
            )
        )

    @classmethod
    def from_csv(
        cls,
        path: str,
        parse_options: CsvParseOptions | None = None,
        io_config: IOConfig | None = None,
        multithreaded_io: bool | None = None,
    ) -> Schema:
        return Schema._from_pyschema(
            _read_csv_schema(
                uri=path,
                parse_options=parse_options,
                io_config=io_config,
                multithreaded_io=multithreaded_io,
            )
        )

    @classmethod
    def from_json(
        cls,
        path: str,
        parse_options: JsonParseOptions | None = None,
        io_config: IOConfig | None = None,
        multithreaded_io: bool | None = None,
    ) -> Schema:
        return Schema._from_pyschema(
            _read_json_schema(
                uri=path,
                parse_options=parse_options,
                io_config=io_config,
                multithreaded_io=multithreaded_io,
            )
        )



from __future__ import annotations

import uuid

import pytest

import daft
from daft import DataFrame
from daft.recordbatch import MicroPartition

NUM_ROWS = 1_000_000


# Perform if/else against two int64 columns, selecting exactly half of the first and half of the second
def generate_int64_params() -> tuple[dict, daft.Expression, list]:
    return (
        {"lhs": [0 for _ in range(NUM_ROWS)], "rhs": [1 for _ in range(NUM_ROWS)], "pred": list(range(NUM_ROWS))},
        (daft.col("pred") < NUM_ROWS // 2).if_else(daft.col("lhs"), daft.col("rhs")),
        [0 for _ in range(NUM_ROWS // 2)] + [1 for _ in range(NUM_ROWS // 2)],
    )


# Perform if/else against two int64 columns, selecting exactly half of the first and half of the second
# This differs from `generate_int64_params` in that the columns and predicate can contain nulls
def generate_int64_with_nulls_params() -> tuple[dict, daft.Expression, list]:
    lhs = [0 if i % 2 == 0 else None for i in range(NUM_ROWS)]
    rhs = [1 if i % 2 == 0 else None for i in range(NUM_ROWS)]
    pred = [i if i % 3 == 0 else None for i in range(NUM_ROWS)]
    expected = [x if p is not None else None for x, p in zip(lhs[: NUM_ROWS // 2] + rhs[NUM_ROWS // 2 :], pred)]
    return (
        {"lhs": lhs, "rhs": rhs, "pred": pred},
        (daft.col("pred") < NUM_ROWS // 2).if_else(daft.col("lhs"), daft.col("rhs")),
        expected,
    )


# Perform if/else against two int64 columns, where the first column is broadcasted
def generate_int64_broadcast_lhs_params() -> tuple[dict, daft.Expression, list]:
    return (
        {"rhs": [1 for _ in range(NUM_ROWS)], "pred": list(range(NUM_ROWS))},
        (daft.col("pred") < NUM_ROWS // 2).if_else(daft.lit(0), daft.col("rhs")),
        [0 for _ in range(NUM_ROWS // 2)] + [1 for _ in range(NUM_ROWS // 2)],
    )


# Perform if/else against two int64 columns, where the first column is broadcasted and predicate has nulls
def generate_int64_broadcast_lhs_with_nulls_params() -> tuple[dict, daft.Expression, list]:
    pred = [i if i % 2 == 0 else None for i in range(NUM_ROWS)]
    expected = [
        x if p is not None else None
        for x, p in zip([0 for _ in range(NUM_ROWS // 2)] + [1 for _ in range(NUM_ROWS // 2)], pred)
    ]
    return (
        {"rhs": [1 for _ in range(NUM_ROWS)], "pred": pred},
        (daft.col("pred") < NUM_ROWS // 2).if_else(daft.lit(0), daft.col("rhs")),
        expected,
    )


# Perform if/else against two string columns, selecting exactly half of the first and half of the second
def generate_string_params() -> tuple[dict, daft.Expression, list]:
    lhs = [str(uuid.uuid4()) for _ in range(NUM_ROWS)]
    rhs = [str(uuid.uuid4()) for _ in range(NUM_ROWS)]
    return (
        {"lhs": lhs, "rhs": rhs, "pred": list(range(NUM_ROWS))},
        (daft.col("pred") < NUM_ROWS // 2).if_else(daft.col("lhs"), daft.col("rhs")),
        lhs[: NUM_ROWS // 2] + rhs[NUM_ROWS // 2 :],
    )


# Perform if/else against two list columns, selecting exactly half of the first and half of the second
def generate_list_params() -> tuple[dict, daft.Expression, list]:
    lhs = [[0 for _ in range(5)] for _ in range(NUM_ROWS)]
    rhs = [[1 for _ in range(5)] for _ in range(NUM_ROWS)]
    return (
        {"lhs": lhs, "rhs": rhs, "pred": list(range(NUM_ROWS))},
        (daft.col("pred") < NUM_ROWS // 2).if_else(daft.col("lhs"), daft.col("rhs")),
        lhs[: NUM_ROWS // 2] + rhs[NUM_ROWS // 2 :],
    )


@pytest.mark.benchmark(group="if_else")
@pytest.mark.parametrize(
    "test_data_generator",
    [
        pytest.param(
            generate_int64_params,
            id="int64",
        ),
        pytest.param(
            generate_int64_with_nulls_params,
            id="int64-with-nulls",
        ),
        pytest.param(
            generate_int64_broadcast_lhs_params,
            id="int64-broadcast-lhs",
        ),
        pytest.param(
            generate_int64_broadcast_lhs_with_nulls_params,
            id="int64-broadcast-lhs-with-nulls",
        ),
        pytest.param(
            generate_string_params,
            id="string",
        ),
        pytest.param(
            generate_list_params,
            id="list",
        ),
    ],
)
def test_if_else(test_data_generator, benchmark) -> None:
    """If_else between NUM_ROWS values."""
    data, expr, expected = test_data_generator()
    table = MicroPartition.from_pydict(data)

    def bench_if_else() -> DataFrame:
        return table.eval_expression_list([expr.alias("result")])

    result = benchmark(bench_if_else)
    assert result.to_pydict()["result"] == expected



from __future__ import annotations

import os
import tempfile

import daft
from daft import col, lit


def test_url_download():
    df = daft.from_pydict({"one": [1]})  # just have a single row, doesn't matter what it is
    url = "https://daft-public-data.s3.us-west-2.amazonaws.com/test_fixtures/small_images/rickroll0.jpg"

    # download one
    df_actual = daft.sql(f"SELECT url_download('{url}') as downloaded FROM df").collect().to_pydict()
    df_expect = df.select(lit(url).url.download().alias("downloaded")).collect().to_pydict()

    assert df_actual == df_expect


def test_url_download_multi():
    df = daft.from_pydict(
        {
            "urls": [
                "https://daft-public-data.s3.us-west-2.amazonaws.com/test_fixtures/small_images/rickroll0.jpg",
                "https://daft-public-data.s3.us-west-2.amazonaws.com/test_fixtures/small_images/rickroll1.jpg",
            ]
        }
    )

    actual = (
        daft.sql(
            """
        SELECT
            url_download(urls) as downloaded,
            url_download(urls, max_connections=>1) as downloaded_single_conn,
            url_download(urls, on_error=>'null') as downloaded_ignore_errors
        FROM df
        """
        )
        .collect()
        .to_pydict()
    )

    expected = (
        df.select(
            col("urls").url.download().alias("downloaded"),
            col("urls").url.download(max_connections=1).alias("downloaded_single_conn"),
            col("urls").url.download(on_error="null").alias("downloaded_ignore_errors"),
        )
        .collect()
        .to_pydict()
    )

    assert actual == expected


def test_url_upload():
    with tempfile.TemporaryDirectory() as tmp_dir:
        df = daft.from_pydict(
            {
                "data": [b"test1", b"test2"],
                "paths": [
                    os.path.join(tmp_dir, "test1.txt"),
                    os.path.join(tmp_dir, "test2.txt"),
                ],
            }
        )

        actual = (
            daft.sql(
                """
            SELECT
                url_upload(data, paths) as uploaded,
                url_upload(data, paths, max_connections=>1) as uploaded_single_conn,
                url_upload(data, paths, on_error=>'null') as uploaded_ignore_errors
            FROM df
            """
            )
            .collect()
            .to_pydict()
        )

        expected = (
            df.select(
                col("data").url.upload(daft.col("paths")).alias("uploaded"),
                col("data").url.upload(daft.col("paths"), max_connections=1).alias("uploaded_single_conn"),
                col("data").url.upload(daft.col("paths"), on_error="null").alias("uploaded_ignore_errors"),
            )
            .collect()
            .to_pydict()
        )

        assert actual == expected

        # Verify files were created
        assert os.path.exists(os.path.join(tmp_dir, "test1.txt"))
        assert os.path.exists(os.path.join(tmp_dir, "test2.txt"))



from __future__ import annotations

import functools
import math
from typing import Any

import pandas as pd
import pytest

import daft


def grouped_stddev(rows) -> tuple[list[Any], list[Any]]:
    map = {}
    for key, data in rows:
        if key not in map:
            map[key] = []
        map[key].append(data)

    keys = []
    stddevs = []
    for key, nums in map.items():
        keys.append(key)
        stddevs.append(stddev(nums))

    return keys, stddevs


def stddev(nums) -> float:
    nums = [num for num in nums if num is not None]

    if not nums:
        return 0.0
    sum_: float = sum(nums)
    count = len(nums)
    mean = sum_ / count

    squared_sums = functools.reduce(lambda acc, num: acc + (num - mean) ** 2, nums, 0)
    stddev = math.sqrt(squared_sums / count)
    return stddev


TESTS = [
    [nums := [0], stddev(nums)],
    [nums := [1], stddev(nums)],
    [nums := [0, 1, 2], stddev(nums)],
    [nums := [100, 100, 100], stddev(nums)],
    [nums := [None, 100, None], stddev(nums)],
    [nums := [None] * 10 + [100], stddev(nums)],
]


@pytest.mark.parametrize("data_and_expected", TESTS)
def test_stddev_with_single_partition(data_and_expected, with_morsel_size):
    data, expected = data_and_expected
    df = daft.from_pydict({"a": data})
    result = df.agg(daft.col("a").stddev()).collect()
    rows = result.iter_rows()
    stddev = next(rows)
    try:
        next(rows)
        assert False
    except StopIteration:
        pass

    assert stddev["a"] == expected


@pytest.mark.parametrize("data_and_expected", TESTS)
def test_stddev_with_multiple_partitions(data_and_expected, with_morsel_size):
    data, expected = data_and_expected
    df = daft.from_pydict({"a": data}).into_partitions(2)
    result = df.agg(daft.col("a").stddev()).collect()
    rows = result.iter_rows()
    stddev = next(rows)
    try:
        next(rows)
        assert False
    except StopIteration:
        pass

    assert stddev["a"] == expected


GROUPED_TESTS = [
    [rows := [("k1", 0), ("k2", 1), ("k1", 1)], *grouped_stddev(rows)],
    [rows := [("k0", 100), ("k1", 100), ("k2", 100)], *grouped_stddev(rows)],
    [rows := [("k0", 100), ("k0", 100), ("k0", 100)], *grouped_stddev(rows)],
    [rows := [("k0", 0), ("k0", 1), ("k0", 2)], *grouped_stddev(rows)],
    [rows := [("k0", None), ("k0", None), ("k0", 100)], *grouped_stddev(rows)],
]


def unzip_rows(rows: list) -> tuple[list, list]:
    keys = []
    nums = []
    for key, data in rows:
        keys.append(key)
        nums.append(data)
    return keys, nums


@pytest.mark.parametrize("data_and_expected", GROUPED_TESTS)
def test_grouped_stddev_with_single_partition(data_and_expected, with_morsel_size):
    nums, expected_keys, expected_stddevs = data_and_expected
    expected_df = daft.from_pydict({"keys": expected_keys, "data": expected_stddevs})
    keys, data = unzip_rows(nums)
    df = daft.from_pydict({"keys": keys, "data": data})
    result_df = df.groupby("keys").agg(daft.col("data").stddev()).collect()

    result = result_df.to_pydict()
    expected = expected_df.to_pydict()

    pd.testing.assert_series_equal(
        pd.Series(result["keys"]).sort_values(),
        pd.Series(expected["keys"]).sort_values(),
        check_index=False,
    )
    pd.testing.assert_series_equal(
        pd.Series(result["data"]).sort_values(),
        pd.Series(expected["data"]).sort_values(),
        check_index=False,
    )


@pytest.mark.parametrize("data_and_expected", GROUPED_TESTS)
def test_grouped_stddev_with_multiple_partitions(data_and_expected, with_morsel_size):
    nums, expected_keys, expected_stddevs = data_and_expected
    expected_df = daft.from_pydict({"keys": expected_keys, "data": expected_stddevs})
    keys, data = unzip_rows(nums)
    df = daft.from_pydict({"keys": keys, "data": data}).into_partitions(2)
    result_df = df.groupby("keys").agg(daft.col("data").stddev()).collect()

    result = result_df.to_pydict()
    expected = expected_df.to_pydict()

    pd.testing.assert_series_equal(
        pd.Series(result["keys"]).sort_values(),
        pd.Series(expected["keys"]).sort_values(),
        check_index=False,
    )
    pd.testing.assert_series_equal(
        pd.Series(result["data"]).sort_values(),
        pd.Series(expected["data"]).sort_values(),
        check_index=False,
    )



from __future__ import annotations

import daft
from daft import coalesce, col, lit


def assert_eq(actual, expect):
    """Asserts two dataframes are equal for tests."""
    assert actual.to_pydict() == expect.to_pydict()


def test_coalesce_basic():
    df = daft.from_pydict(
        {
            "a": [None, 1, None, 3],
            "b": [2, None, 4, 5],
            "c": [6, 7, None, 9],
        }
    )
    #
    # single arg
    expect = df.select(coalesce(col("a")).alias("result"))
    actual = daft.sql("select coalesce(a) as result from df")
    assert_eq(actual, expect)
    #
    # two args
    expect = df.select(coalesce(col("a"), col("b")).alias("result"))
    actual = daft.sql("select coalesce(a, b) as result from df")
    assert_eq(actual, expect)
    #
    # three args
    expect = df.select(coalesce(col("a"), col("b"), col("c")).alias("result"))
    actual = daft.sql("select coalesce(a, b, c) as result from df")
    assert_eq(actual, expect)


def test_coalesce_typing():
    """Test coalesce with arguments of different but compatible types."""
    df = daft.from_pydict(
        {
            "int_col": [None, 1, None, 3],
            "float_col": [2.5, None, 4.5, None],
            "str_col": ["6", "7", None, "9"],
        }
    )
    # ok, these are coercible
    # https://github.com/Eventual-Inc/Daft/issues/3752
    daft.sql("SELECT coalesce(1::bigint, 2::int) FROM df").collect()
    daft.sql("SELECT coalesce(1::int, 2::bigint) FROM df").collect()
    #
    # comparable types with minimal common supertype in first position
    actual = daft.sql("select coalesce(float_col, int_col) as result from df")
    expect = df.select(coalesce(col("float_col"), col("int_col")).alias("result"))
    assert_eq(actual, expect)
    #
    # comparable types with minimal common supertype NOT in first position
    actual = daft.sql("select coalesce(int_col, float_col) as result from df")
    expect = df.select(coalesce(col("int_col"), col("float_col")).alias("result"))
    assert_eq(actual, expect)
    #
    # ok in Daft, but not in standard SQL (str and int not comparable).
    #   I believe we should keep as-is because changes to the minimal common supertype logic
    #   would impact all other operators using get_super_type and it is currently quite forgiving.
    daft.sql("select coalesce(int_col, float_col, str_col) as result from df").collect()


def test_coalesce_argument_order():
    df = daft.from_pydict(
        {
            "a": [None, 1, None, 3, 0],
            "b": [2, None, 4, None, 1],
        }
    )
    #
    # order: a, b
    expect_ab = df.select(coalesce(col("a"), col("b")).alias("result"))
    actual_ab = daft.sql("select coalesce(a, b) as result from df")
    assert_eq(actual_ab, expect_ab)
    #
    # order: b, a
    expect_ba = df.select(coalesce(col("b"), col("a")).alias("result"))
    actual_ba = daft.sql("select coalesce(b, a) as result from df")
    assert_eq(actual_ba, expect_ba)
    #
    # values should be different due to different argument order
    assert expect_ab.to_pydict() != expect_ba.to_pydict()


def test_coalesce_literals():
    df = daft.from_pydict({"dummy": [1, 2, 3, 4]})
    #
    # all literals
    actual = daft.sql("select coalesce(null, 1, 2) as result from df")
    expect = daft.from_pydict({"result": [1, 1, 1, 1]})
    assert_eq(actual, expect)
    #
    # mix of column and literals
    actual = daft.sql("select coalesce(dummy, 99) as result from df")
    expect = df.select(coalesce(col("dummy"), lit(99)).alias("result"))
    assert_eq(actual, expect)


def test_coalesce_with_nulls():
    df = daft.from_pydict({"col1": [None, None, None, None]})
    #
    # all null arguments
    actual = daft.sql("select coalesce(null, null, null) as result from df")
    expect = daft.from_pydict({"result": [None, None, None, None]})
    assert_eq(actual, expect)
    #
    # null last
    actual = daft.sql("select coalesce(col1, null) as result from df")
    expect = df.select(coalesce(col("col1")).alias("result"))
    assert_eq(actual, expect)
    #
    # null first
    actual = daft.sql("select coalesce(null, col1) as result from df")
    expect = df.select(coalesce(col("col1")).alias("result"))
    assert_eq(actual, expect)



from __future__ import annotations

import pytest

import daft
from daft.pyspark import SparkSession


@pytest.fixture(params=["local_spark", "ray_spark"], scope="session")
def spark_session(request):
    return request.getfixturevalue(request.param)


@pytest.fixture(scope="session")
def local_spark():
    session = SparkSession.builder.appName("DaftConfigTest").local().getOrCreate()
    yield session
    session.stop()


@pytest.fixture(scope="session")
def ray_spark():
    session = SparkSession.builder.appName("DaftConfigTest").remote("ray://localhost:10001").getOrCreate()
    yield session
    session.stop()


@pytest.fixture(scope="function")
def make_spark_df(spark_session):
    def make_df(data):
        fields = [name for name in data]
        rows = list(zip(*[data[name] for name in fields]))
        return spark_session.createDataFrame(rows, fields)

    yield make_df


@pytest.fixture(scope="function")
def spark_to_daft():
    def inner(spark_df):
        return daft.from_pandas(spark_df.toPandas())

    yield inner



from __future__ import annotations

import pathlib
import sys
import uuid

import pandas as pd
import pytest

import daft
from daft.expressions import col
from tests.conftest import assert_df_equals


def _get_filename():
    name = str(uuid.uuid4())

    # Inject colons into the name if not windows
    if sys.platform != "win32":
        name += ":foo:bar"

    return name


@pytest.fixture(scope="function")
def files(tmpdir) -> list[str]:
    filepaths = [pathlib.Path(tmpdir) / _get_filename() for _ in range(10)]
    for fp in filepaths:
        fp.write_bytes(fp.name.encode("utf-8"))
    return filepaths


def test_download(files):
    # Run it twice to ensure runtime works
    for _ in range(2):
        df = daft.from_pydict({"filenames": [str(f) for f in files]})
        df = df.with_column("bytes", col("filenames").url.download())
        pd_df = pd.DataFrame.from_dict({"filenames": [str(f) for f in files]})
        pd_df["bytes"] = pd.Series([pathlib.Path(fn).read_bytes() for fn in files])
        assert_df_equals(df.to_pandas(), pd_df, sort_key="filenames")


def test_download_with_none(files):
    data = {"id": list(range(len(files) * 2)), "filenames": [str(f) for f in files] + [None for _ in range(len(files))]}
    # Run it twice to ensure runtime works
    for _ in range(2):
        df = daft.from_pydict(data)
        df = df.with_column("bytes", col("filenames").url.download())
        pd_df = pd.DataFrame.from_dict(data)
        pd_df["bytes"] = pd.Series([pathlib.Path(fn).read_bytes() if fn is not None else None for fn in files])
        assert_df_equals(df.to_pandas(), pd_df, sort_key="id")


def test_download_with_missing_urls(files):
    data = {
        "id": list(range(len(files) * 2)),
        "filenames": [str(f) for f in files] + [str(uuid.uuid4()) for _ in range(len(files))],
    }
    # Run it twice to ensure runtime works
    for _ in range(2):
        df = daft.from_pydict(data)
        df = df.with_column(
            "bytes",
            col("filenames").url.download(
                on_error="null",
            ),
        )
        pd_df = pd.DataFrame.from_dict(data)
        pd_df["bytes"] = pd.Series(
            [pathlib.Path(fn).read_bytes() if pathlib.Path(fn).exists() else None for fn in files]
        )
        assert_df_equals(df.to_pandas(), pd_df, sort_key="id")


def test_download_with_missing_urls_reraise_errors(files):
    data = {
        "id": list(range(len(files) * 2)),
        "filenames": [str(f) for f in files] + [str(uuid.uuid4()) for _ in range(len(files))],
    }
    # Run it twice to ensure runtime works
    for _ in range(2):
        df = daft.from_pydict(data)
        df = df.with_column(
            "bytes",
            col("filenames").url.download(
                on_error="raise",
            ),
        )
        # TODO: Change to a FileNotFound Error

        with pytest.raises(FileNotFoundError):
            df.collect()


def test_download_with_duplicate_urls(files):
    data = {
        "id": list(range(len(files) * 2)),
        "filenames": [str(f) for f in files] * 2,
    }
    # Run it twice to ensure runtime works
    for _ in range(2):
        df = daft.from_pydict(data)
        df = df.with_column("bytes", col("filenames").url.download())
        pd_df = pd.DataFrame.from_dict(data)
        pd_df["bytes"] = pd.Series(
            [pathlib.Path(fn).read_bytes() if pathlib.Path(fn).exists() else None for fn in files * 2]
        )
        assert_df_equals(df.to_pandas(), pd_df, sort_key="id")



from __future__ import annotations

import pytest

import daft
from daft import DataType
from daft.recordbatch import MicroPartition

MODE_TO_NUM_CHANNELS = {
    "L": 1,
    "LA": 2,
    "RGB": 3,
    "RGBA": 4,
    "L16": 1,
    "LA16": 2,
    "RGB16": 3,
    "RGBA16": 4,
    "RGB32F": 3,
    "RGBA32F": 4,
}


@pytest.mark.parametrize("out_mode", ["L", "LA", "RGB", "RGBA"])
def test_image_to_mode_mixed_shape(mixed_shape_data_fixture, out_mode):
    table = MicroPartition.from_pydict({"images": mixed_shape_data_fixture})
    result = table.eval_expression_list([daft.col("images").image.to_mode(out_mode)])
    assert result.schema()["images"].dtype == DataType.image(out_mode)
    result = result.to_pydict()
    assert [x.shape[2] == MODE_TO_NUM_CHANNELS[out_mode] if x is not None else None for x in result["images"]]


@pytest.mark.parametrize("out_mode", ["L", "LA", "RGB", "RGBA"])
def test_image_crop_fixed_shape(fixed_shape_data_fixture, out_mode):
    table = MicroPartition.from_pydict({"images": fixed_shape_data_fixture})
    result = table.eval_expression_list([daft.col("images").image.to_mode(out_mode)])
    assert result.schema()["images"].dtype == DataType.image(out_mode, 3, 4)
    result = result.to_pydict()
    assert [x.shape == (3, 4, MODE_TO_NUM_CHANNELS[out_mode]) if x is not None else None for x in result["images"]]



from __future__ import annotations

from daft.expressions import col
from daft.recordbatch import MicroPartition


def test_utf8_lstrip():
    table = MicroPartition.from_pydict({"col": ["\ta\t", None, "\nb\n", "\vc\t", "\td ", "e"]})
    result = table.eval_expression_list([col("col").str.lstrip()])
    assert result.to_pydict() == {"col": ["a\t", None, "b\n", "c\t", "d ", "e"]}



# ruff: noqa: I002
# isort: dont-add-import: from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional, Union

from daft import context
from daft.api_annotations import PublicAPI
from daft.daft import IOConfig, ScanOperatorHandle, StorageConfig
from daft.dataframe import DataFrame
from daft.logical.builder import LogicalPlanBuilder

if TYPE_CHECKING:
    from pyiceberg.table import Table as PyIcebergTable


def _convert_iceberg_file_io_properties_to_io_config(props: dict[str, Any]) -> Optional[IOConfig]:
    """Property keys defined here: https://github.com/apache/iceberg-python/blob/main/pyiceberg/io/__init__.py."""
    from daft.io import AzureConfig, GCSConfig, IOConfig, S3Config

    any_props_set = False

    def get_first_property_value(*property_names: str) -> Optional[Any]:
        for property_name in property_names:
            if property_value := props.get(property_name):
                nonlocal any_props_set
                any_props_set = True
                return property_value
        return None

    io_config = IOConfig(
        s3=S3Config(
            endpoint_url=get_first_property_value("s3.endpoint"),
            region_name=get_first_property_value("s3.region", "client.region"),
            key_id=get_first_property_value("s3.access-key-id", "client.access-key-id"),
            access_key=get_first_property_value("s3.secret-access-key", "client.secret-access-key"),
            session_token=get_first_property_value("s3.session-token", "client.session-token"),
        ),
        azure=AzureConfig(
            storage_account=get_first_property_value("adls.account-name", "adlfs.account-name"),
            access_key=get_first_property_value("adls.account-key", "adlfs.account-key"),
            sas_token=get_first_property_value("adls.sas-token", "adlfs.sas-token"),
            tenant_id=get_first_property_value("adls.tenant-id", "adlfs.tenant-id"),
            client_id=get_first_property_value("adls.client-id", "adlfs.client-id"),
            client_secret=get_first_property_value("adls.client-secret", "adlfs.client-secret"),
        ),
        gcs=GCSConfig(
            project_id=get_first_property_value("gcs.project-id"),
            token=get_first_property_value("gcs.oauth2.token"),
        ),
    )

    return io_config if any_props_set else None


@PublicAPI
def read_iceberg(
    table: Union[str, "PyIcebergTable"],
    snapshot_id: Optional[int] = None,
    io_config: Optional[IOConfig] = None,
) -> DataFrame:
    """Create a DataFrame from an Iceberg table.

    Args:
        table (str or pyiceberg.table.Table): [PyIceberg Table](https://py.iceberg.apache.org/reference/pyiceberg/table/#pyiceberg.table.Table) created using the PyIceberg library
        snapshot_id (int, optional): Snapshot ID of the table to query
        io_config (IOConfig, optional): A custom IOConfig to use when accessing Iceberg object storage data. If provided, configurations set in `table` are ignored.

    Returns:
        DataFrame: a DataFrame with the schema converted from the specified Iceberg table

    Note:
        This function requires the use of [PyIceberg](https://py.iceberg.apache.org/), which is the Apache Iceberg's
        official project for Python.

    Examples:
        >>> import pyiceberg
        >>>
        >>> table = pyiceberg.Table(...)
        >>> df = daft.read_iceberg(table)
        >>>
        >>> # Filters on this dataframe can now be pushed into
        >>> # the read operation from Iceberg
        >>> df = df.where(df["foo"] > 5)
        >>> df.show()

    """
    from pyiceberg.table import StaticTable

    from daft.iceberg.iceberg_scan import IcebergScanOperator

    # support for read_iceberg('path/to/metadata.json')
    if isinstance(table, str):
        table = StaticTable.from_metadata(metadata_location=table)

    io_config = (
        _convert_iceberg_file_io_properties_to_io_config(table.io.properties) if io_config is None else io_config
    )
    io_config = context.get_context().daft_planning_config.default_io_config if io_config is None else io_config

    multithreaded_io = context.get_context().get_or_create_runner().name != "ray"
    storage_config = StorageConfig(multithreaded_io, io_config)

    iceberg_operator = IcebergScanOperator(table, snapshot_id=snapshot_id, storage_config=storage_config)

    handle = ScanOperatorHandle.from_python_scan_operator(iceberg_operator)
    builder = LogicalPlanBuilder.from_tabular_scan(scan_operator=handle)
    return DataFrame(builder)



from __future__ import annotations

import asyncio
import logging
from dataclasses import dataclass
from typing import TYPE_CHECKING

from daft.daft import (
    DistributedPhysicalPlan,
    DistributedPhysicalPlanRunner,
    LocalPhysicalPlan,
    NativeExecutor,
    PyDaftExecutionConfig,
    RayPartitionRef,
    RaySwordfishTask,
    RaySwordfishWorker,
    set_compute_runtime_num_worker_threads,
)
from daft.recordbatch.micropartition import MicroPartition
from daft.runners.constants import (
    MAX_SWORDFISH_ACTOR_RESTARTS,
    MAX_SWORDFISH_ACTOR_TASK_RETRIES,
)
from daft.runners.partitioning import (
    PartitionMetadata,
    PartitionSet,
)

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator, AsyncIterator

    from daft.runners.ray_runner import RayMaterializedResult

try:
    import ray
except ImportError:
    raise

logger = logging.getLogger(__name__)


@ray.remote(
    max_restarts=MAX_SWORDFISH_ACTOR_RESTARTS,
    max_task_retries=MAX_SWORDFISH_ACTOR_TASK_RETRIES,
)
class RaySwordfishActor:
    """RaySwordfishActor is a ray actor that runs local physical plans on swordfish.

    It is a stateless, async actor, and can run multiple plans concurrently and is able to retry itself and it's tasks.
    """

    def __init__(self, num_worker_threads: int) -> None:
        # Configure the number of worker threads for swordfish, according to the number of CPUs visible to ray.
        set_compute_runtime_num_worker_threads(num_worker_threads)
        self.native_executor = NativeExecutor()

    async def run_plan(
        self,
        plan: LocalPhysicalPlan,
        config: PyDaftExecutionConfig,
        psets: dict[str, list[ray.ObjectRef]],
        context: dict[str, str] | None,
    ) -> AsyncGenerator[MicroPartition | list[PartitionMetadata], None]:
        """Run a plan on swordfish and yield partitions."""
        psets = {k: await asyncio.gather(*v) for k, v in psets.items()}
        psets_mp = {k: [v._micropartition for v in v] for k, v in psets.items()}

        metas = []
        async for partition in self.native_executor.run_async(plan, psets_mp, config, None, context):
            if partition is None:
                break
            mp = MicroPartition._from_pymicropartition(partition)
            metas.append(PartitionMetadata.from_table(mp))
            yield mp
        yield metas


@dataclass
class RaySwordfishTaskHandle:
    """RaySwordfishTaskHandle is a handle to a task that is running on a swordfish actor.

    It is used to asynchronously get the result of the task, cancel the task, and perform any post-task cleanup.
    """

    result_handle: ray.ObjectRef
    actor_handle: ray.actor.ActorHandle
    task: asyncio.Task[list[RayPartitionRef]] | None = None

    async def _get_result(self) -> list[RayPartitionRef]:
        await self.result_handle.completed()
        results = [result for result in self.result_handle]
        metadatas_ref = results.pop()

        metadatas = await metadatas_ref
        assert len(results) == len(metadatas)

        res = [
            RayPartitionRef(result, metadata.num_rows, metadata.size_bytes)
            for result, metadata in zip(results, metadatas)
        ]
        return res

    async def get_result(self) -> list[RayPartitionRef]:
        self.task = asyncio.create_task(self._get_result())
        return await self.task

    def cancel(self) -> None:
        if self.task:
            self.task.cancel()
        ray.cancel(self.result_handle)


class RaySwordfishActorHandle:
    """RaySwordfishWorkerHandle is a wrapper around a ray swordfish actor.

    It is used to submit tasks to the worker and keep track of the worker's node id, handle, number of cpus, total and available memory.
    """

    def __init__(
        self,
        actor_handle: ray.actor.ActorHandle,
    ):
        self.actor_handle = actor_handle

    def submit_task(self, task: RaySwordfishTask) -> RaySwordfishTaskHandle:
        psets = {k: [v.object_ref for v in v] for k, v in task.psets().items()}
        result_handle = self.actor_handle.run_plan.options(name=task.name()).remote(
            task.plan(), task.config(), psets, task.context()
        )
        return RaySwordfishTaskHandle(
            result_handle,
            self.actor_handle,
        )

    def shutdown(self) -> None:
        ray.kill(self.actor_handle)


def start_ray_workers() -> list[RaySwordfishWorker]:
    handles = []
    for node in ray.nodes():
        if (
            "Resources" in node
            and "CPU" in node["Resources"]
            and "memory" in node["Resources"]
            and node["Resources"]["CPU"] > 0
            and node["Resources"]["memory"] > 0
        ):
            actor = RaySwordfishActor.options(  # type: ignore
                scheduling_strategy=ray.util.scheduling_strategies.NodeAffinitySchedulingStrategy(
                    node_id=node["NodeID"],
                    soft=False,
                ),
            ).remote(num_worker_threads=int(node["Resources"]["CPU"]))
            actor_handle = RaySwordfishActorHandle(actor)
            handles.append(
                RaySwordfishWorker(
                    node["NodeID"],
                    actor_handle,
                    int(node["Resources"]["CPU"]),
                    int(node["Resources"].get("GPU", 0)),
                    int(node["Resources"]["memory"]),
                )
            )

    return handles


@ray.remote(
    num_cpus=0,
)
class FlotillaPlanRunner:
    def __init__(self) -> None:
        self.curr_plans: dict[str, DistributedPhysicalPlan] = {}
        self.curr_result_gens: dict[str, AsyncIterator[tuple[ray.ObjectRef, int, int]]] = {}
        self.plan_runner = DistributedPhysicalPlanRunner()

    def run_plan(
        self,
        plan: DistributedPhysicalPlan,
        partition_sets: dict[str, PartitionSet[ray.ObjectRef]],
    ) -> None:
        psets = {
            k: [RayPartitionRef(v.partition(), v.metadata().num_rows, v.metadata().size_bytes or 0) for v in v.values()]
            for k, v in partition_sets.items()
        }
        self.curr_plans[plan.id()] = plan
        self.curr_result_gens[plan.id()] = self.plan_runner.run_plan(plan, psets)

    async def get_next_partition(self, plan_id: str) -> RayMaterializedResult | None:
        from daft.runners.ray_runner import (
            PartitionMetadataAccessor,
            RayMaterializedResult,
        )

        if plan_id not in self.curr_result_gens:
            raise ValueError(f"Plan {plan_id} not found in FlotillaPlanRunner")

        next_result = await self.curr_result_gens[plan_id].__anext__()
        if next_result is None:
            self.curr_plans.pop(plan_id)
            self.curr_result_gens.pop(plan_id)
            return None

        obj, num_rows, size_bytes = next_result
        metadata_accessor = PartitionMetadataAccessor.from_metadata_list([PartitionMetadata(num_rows, size_bytes)])
        materialized_result = RayMaterializedResult(
            partition=obj,
            metadatas=metadata_accessor,
            metadata_idx=0,
        )
        return materialized_result



from __future__ import annotations

import uuid

import pytest

from daft.series import Series


@pytest.mark.benchmark(group="concat")
def test_string_concat(benchmark) -> None:
    NUM_ROWS = 100_000
    data = Series.from_pylist([str(uuid.uuid4()) for _ in range(NUM_ROWS)])
    to_concat = [data] * 100

    def bench_concat() -> Series:
        return Series.concat(to_concat)

    benchmark(bench_concat)



from __future__ import annotations

import pyarrow as pa
import pytest

from daft.datatype import DataType
from daft.series import Series
from tests.series import ARROW_FLOAT_TYPES, ARROW_INT_TYPES


@pytest.mark.parametrize("dtype", ARROW_INT_TYPES + ARROW_FLOAT_TYPES)
def test_series_numeric_abs(dtype) -> None:
    if pa.types.is_unsigned_integer(dtype):
        pydata = list(range(0, 10))
    else:
        pydata = list(range(-10, 10))

    data = pa.array(pydata, dtype)

    s = Series.from_arrow(data)

    assert s.datatype() == DataType.from_arrow_type(dtype)

    abs_s = abs(s)

    assert abs_s.datatype() == DataType.from_arrow_type(dtype)

    assert abs_s.to_pylist() == list(map(abs, pydata))


def test_table_abs_bad_input() -> None:
    series = Series.from_pylist(["a", "b", "c"])

    with pytest.raises(ValueError, match="abs not implemented"):
        abs(series)



from __future__ import annotations

import datetime
import os
import platform
import socket
import time
import urllib
from unittest.mock import MagicMock, patch

import pytest

import daft
from daft.analytics import AnalyticsClient

PUBLISHER_THREAD_SLEEP_INTERVAL_SECONDS = 0.1
MOCK_DATETIME = datetime.datetime(2021, 1, 1, 0, 0, 0)


@pytest.fixture(scope="function")
def mock_analytics() -> tuple[AnalyticsClient, MagicMock]:
    mock_publish = MagicMock()
    client = AnalyticsClient(
        daft.get_version(),
        daft.get_build_type(),
        True,
        publish_payload_function=mock_publish,
        buffer_capacity=1,
    )

    # Patch the publish method
    client._post_segment_track_endpoint = MagicMock()

    return client, mock_publish


@patch("daft.analytics.datetime")
def test_analytics_client_track_import(mock_datetime: MagicMock, mock_analytics: tuple[AnalyticsClient, MagicMock]):
    mock_datetime.datetime.now.return_value = MOCK_DATETIME
    analytics_client, mock_publish = mock_analytics

    # Run track_import
    analytics_client.track_import()

    # Sleep to allow publisher thread to poll events
    time.sleep(PUBLISHER_THREAD_SLEEP_INTERVAL_SECONDS + 0.5)

    mock_publish.assert_called_once_with(
        analytics_client,
        {
            "batch": [
                {
                    "type": "track",
                    "anonymousId": analytics_client._session_key,
                    "event": "Imported Daft",
                    "properties": {
                        "platform": platform.platform(),
                        "python_version": platform.python_version(),
                        "DAFT_ANALYTICS_ENABLED": os.getenv("DAFT_ANALYTICS_ENABLED"),
                    },
                    "timestamp": MOCK_DATETIME.isoformat(),
                    "context": {
                        "app": {
                            "name": "daft",
                            "version": daft.get_version(),
                            "build": daft.get_build_type(),
                        },
                    },
                }
            ],
        },
    )


@patch("urllib.request.urlopen")
def test_analytics_client_timeout(
    mock_urlopen: MagicMock,
):
    mock_urlopen.side_effect = urllib.error.URLError(socket.timeout("Timeout"))
    analytics_client = AnalyticsClient(
        daft.get_version(),
        daft.get_build_type(),
        True,
        buffer_capacity=1,
    )

    # Call `track_import` 5 times and ensure that, after the first timeout, successive calls to `urllib.request.urlopen` are never made
    for _ in range(5):
        analytics_client.track_import()
    mock_urlopen.assert_called_once()


@patch("urllib.request.urlopen")
def test_analytics_client_timeout_2(
    mock_urlopen: MagicMock,
):
    mock_urlopen.return_value.status = 408
    analytics_client = AnalyticsClient(
        daft.get_version(),
        daft.get_build_type(),
        True,
        buffer_capacity=1,
    )
    analytics_client.track_import()
    mock_urlopen.assert_called_once()


@patch("urllib.request.urlopen")
def test_analytics_client_disabled(
    mock_urlopen: MagicMock,
):
    mock_urlopen.side_effect = urllib.error.URLError(socket.timeout("Timeout"))
    analytics_client = AnalyticsClient(
        daft.get_version(),
        daft.get_build_type(),
        False,
        buffer_capacity=1,
    )

    analytics_client.track_import()
    mock_urlopen.assert_not_called()


@patch("daft.analytics.datetime")
def test_analytics_client_track_dataframe_method(
    mock_datetime: MagicMock, mock_analytics: tuple[AnalyticsClient, MagicMock]
):
    mock_datetime.datetime.now.return_value = MOCK_DATETIME
    analytics_client, mock_publish = mock_analytics

    # Run track_df_method_call
    analytics_client.track_df_method_call(
        "foo",
        4.32,
        "err",
    )

    # Sleep to allow publisher thread to poll events
    time.sleep(PUBLISHER_THREAD_SLEEP_INTERVAL_SECONDS + 0.5)

    mock_publish.assert_called_once_with(
        analytics_client,
        {
            "batch": [
                {
                    "type": "track",
                    "anonymousId": analytics_client._session_key,
                    "event": "DataFrame Method Call",
                    "properties": {
                        "method_name": "foo",
                        "duration_seconds": 4.32,
                        "error": "err",
                    },
                    "timestamp": MOCK_DATETIME.isoformat(),
                    "context": {
                        "app": {
                            "name": "daft",
                            "version": daft.get_version(),
                            "build": daft.get_build_type(),
                        },
                    },
                }
            ],
        },
    )



from __future__ import annotations

import pytest

from daft.expressions import col
from daft.functions import format


def test_format_basic(make_df) -> None:
    """Test basic format functionality."""
    data = {"first": ["Alice", "Bob"], "last": ["Smith", "Jones"]}
    df = make_df(data).with_column("greeting", format("Hello {} {}", "first", "last"))

    expected = ["Hello Alice Smith", "Hello Bob Jones"]
    assert df.to_pydict()["greeting"] == expected


def test_format_string_vs_expression_args(make_df) -> None:
    """Test that string column names and col() expressions work the same."""
    data = {"name": ["Alice", "Bob"]}
    df = make_df(data).with_column("greeting", format("Hello {}", "name"))

    expected = ["Hello Alice", "Hello Bob"]
    assert df.to_pydict()["greeting"] == expected


def test_format_no_placeholders(make_df) -> None:
    """Test format with no placeholders returns just the literal string."""
    data = {"name": ["Alice", "Bob"]}
    df = make_df(data).with_column("static", format("Hello World"))

    expected = ["Hello World", "Hello World"]
    assert df.to_pydict()["static"] == expected


def test_format_only_placeholders(make_df) -> None:
    """Test format with only placeholders (no static text)."""
    data = {"first": ["Alice", "Bob"], "last": ["Smith", "Jones"]}
    df = make_df(data).with_column("full", format("{}{}", "first", "last"))

    expected = ["AliceSmith", "BobJones"]
    assert df.to_pydict()["full"] == expected


def test_format_ends_with_placeholder(make_df) -> None:
    """Test that parts[-1] is handled correctly when format string ends with {}."""
    data = {"name": ["Alice", "Bob"]}
    df = make_df(data).with_column("greeting", format("Hello {}", "name"))

    expected = ["Hello Alice", "Hello Bob"]
    assert df.to_pydict()["greeting"] == expected


def test_format_starts_with_placeholder(make_df) -> None:
    """Test format string that starts with a placeholder."""
    data = {"name": ["Alice", "Bob"]}
    df = make_df(data).with_column("greeting", format("{}, welcome!", "name"))

    expected = ["Alice, welcome!", "Bob, welcome!"]
    assert df.to_pydict()["greeting"] == expected


def test_format_multiple_static_parts(make_df) -> None:
    """Test format with multiple static text parts between placeholders."""
    data = {"first": ["Alice", "Bob"], "last": ["Smith", "Jones"]}
    df = make_df(data).with_column("message", format("Dear {} from {}, welcome!", "first", "last"))

    expected = ["Dear Alice from Smith, welcome!", "Dear Bob from Jones, welcome!"]
    assert df.to_pydict()["message"] == expected


def test_format_with_expressions(make_df) -> None:
    """Test format with computed expressions as arguments."""
    data = {"a": [1, 2], "b": [3, 4]}
    df = make_df(data).with_column("result", format("Sum: {}", col("a") + col("b")))

    expected = ["Sum: 4", "Sum: 6"]
    assert df.to_pydict()["result"] == expected


def test_format_placeholder_count_mismatch_too_few_args() -> None:
    """Test error when number of placeholders does not match number of arguments."""
    with pytest.raises(ValueError, match="Format string .* has 2 placeholders but 1 arguments were provided"):
        format("Hello {} {}", "name")


def test_format_empty_string_no_args(make_df) -> None:
    """Test format with empty string and no arguments returns empty string literal."""
    data = {"name": ["Alice", "Bob"]}
    df = make_df(data).with_column("empty", format(""))

    expected = ["", ""]
    assert df.to_pydict()["empty"] == expected



# /// script
# requires-python = ">=3.12"
# dependencies = [
#   "ray[default]",
#   "daft",
# ]
# ///
from __future__ import annotations

import argparse
import asyncio
import json
import os
import pathlib
import time
import uuid
from datetime import datetime
from pathlib import Path
from typing import Callable

from ray.job_submission import JobStatus, JobSubmissionClient

import daft


async def print_logs(logs):
    async for lines in logs:
        print(lines, end="")


async def wait_on_job(logs, timeout_s):
    await asyncio.wait_for(print_logs(logs), timeout=timeout_s)


def run_on_ray(ray_address: str, job_params: dict, timeout_s: int = 1500):
    """Submits a job to run in the Ray cluster."""
    print("Submitting benchmarking job to Ray cluster...")
    print("Parameters:")
    print(job_params)

    client = JobSubmissionClient(address=ray_address)
    job_id = client.submit_job(**job_params)
    print(f"Submitted job: {job_id}")

    try:
        asyncio.run(wait_on_job(client.tail_job_logs(job_id), timeout_s))
    except asyncio.TimeoutError:
        print(f"Job timed out after {timeout_s}s! Stopping job now...")
        client.stop_job(job_id)
        time.sleep(16)

    status = client.get_job_status(job_id)
    assert status.is_terminal(), "Job should have terminated"
    if status != JobStatus.SUCCEEDED:
        job_info = client.get_job_info(job_id)
        raise RuntimeError(f"Job failed with {job_info.error_type} error: {job_info.message}")
    print(f"Job completed with {status}")


def ray_job_params(
    parquet_folder_path: str,
    tpch_qnum: int,
    working_dir: pathlib.Path,
    entrypoint: pathlib.Path,
    runtime_env: dict,
) -> dict:
    return dict(
        submission_id=f"tpch-q{tpch_qnum}-{str(uuid.uuid4())[:4]}",
        entrypoint=f"python3 {entrypoint.relative_to(working_dir)!s} --parquet-folder {parquet_folder_path} --question-number {tpch_qnum}",
        runtime_env={
            "working_dir": str(working_dir),
            **runtime_env,
        },
    )


def get_df_with_parquet_folder(parquet_folder: str) -> Callable[[str], daft.DataFrame]:
    def _get_df(table_name: str) -> daft.DataFrame:
        return daft.read_parquet(os.path.join(parquet_folder, table_name, "*.parquet"))

    return _get_df


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--parquet-folder", help="Path to TPC-H data stored on workers", required=True)
    parser.add_argument("--question-number", help="Question number to run", required=True)
    args = parser.parse_args()

    import answers

    get_df = get_df_with_parquet_folder(args.parquet_folder)
    answer = getattr(answers, f"q{args.question_number}")
    daft_df = answer(get_df)

    info_path = Path("/tmp") / "ray" / "session_latest" / "logs" / "info"
    info_path.mkdir(parents=True, exist_ok=True)

    explain_delta = None
    with open(info_path / f"plan-{args.question_number}.txt", "w") as f:
        explain_start = datetime.now()
        daft_df.explain(show_all=True, file=f, format="mermaid")
        explain_end = datetime.now()
        explain_delta = explain_end - explain_start

    execute_delta = None
    execute_start = datetime.now()
    daft_df.collect()
    execute_end = datetime.now()
    execute_delta = execute_end - execute_start

    with open(info_path / f"stats-{args.question_number}.txt", "w") as f:
        stats = json.dumps(
            {
                "question": args.question_number,
                "planning-time": str(explain_delta),
                "execution-time": str(execute_delta),
            }
        )
        f.write(stats)



from __future__ import annotations

import numpy as np
import pyarrow as pa
import pytest

from daft import Series
from daft.datatype import DataType
from daft.utils import pyarrow_supports_fixed_shape_tensor
from tests.conftest import get_tests_daft_runner_name

ARROW_VERSION = tuple(int(s) for s in pa.__version__.split(".") if s.isnumeric())


@pytest.mark.parametrize("if_true_value", [1, None])
@pytest.mark.parametrize("if_false_value", [0, None])
@pytest.mark.parametrize(
    "if_true_length",
    [1, 3],
)
@pytest.mark.parametrize(
    "if_false_length",
    [1, 3],
)
@pytest.mark.parametrize(
    # Test supercasting.
    "true_false_types",
    [(pa.int64(), pa.int64()), (pa.int64(), pa.int8()), (pa.int8(), pa.int64())],
)
def test_series_if_else_numeric(
    if_true_value,
    if_false_value,
    if_true_length,
    if_false_length,
    true_false_types,
) -> None:
    true_type, false_type = true_false_types

    if_true_series = Series.from_arrow(pa.array([if_true_value] * if_true_length, type=true_type))
    if_false_series = Series.from_arrow(pa.array([if_false_value] * if_false_length, type=false_type))
    predicate_series = Series.from_arrow(pa.array([True, False, None]))
    result = predicate_series.if_else(if_true_series, if_false_series)
    assert result.datatype() == DataType.int64()
    assert result.to_pylist() == [if_true_value, if_false_value, None]


@pytest.mark.parametrize("if_true_value", [object(), None])
@pytest.mark.parametrize("if_false_value", [object(), None])
@pytest.mark.parametrize(
    "if_true_length",
    [1, 3],
)
@pytest.mark.parametrize(
    "if_false_length",
    [1, 3],
)
def test_series_if_else_pyobj(
    if_true_value,
    if_false_value,
    if_true_length,
    if_false_length,
) -> None:
    if_true_series = Series.from_pylist([if_true_value] * if_true_length)
    if_false_series = Series.from_pylist([if_false_value] * if_false_length)
    predicate_series = Series.from_arrow(pa.array([True, False, None]))
    result = predicate_series.if_else(if_true_series, if_false_series)
    assert result.to_pylist() == [if_true_value, if_false_value, None]


@pytest.mark.parametrize("if_true_value", ["1", None])
@pytest.mark.parametrize("if_false_value", ["0", None])
@pytest.mark.parametrize(
    "if_true_length",
    [1, 3],
)
@pytest.mark.parametrize(
    "if_false_length",
    [1, 3],
)
def test_series_if_else_string(
    if_true_value,
    if_false_value,
    if_true_length,
    if_false_length,
) -> None:
    if_true_series = Series.from_arrow(pa.array([if_true_value] * if_true_length, type=pa.string()))
    if_false_series = Series.from_arrow(pa.array([if_false_value] * if_false_length, type=pa.string()))
    predicate_series = Series.from_arrow(pa.array([True, False, None]))
    result = predicate_series.if_else(if_true_series, if_false_series)
    assert result.datatype() == DataType.string()
    assert result.to_pylist() == [if_true_value, if_false_value, None]


@pytest.mark.parametrize("if_true_value", [True, None])
@pytest.mark.parametrize("if_false_value", [None, None])
@pytest.mark.parametrize(
    "if_true_length",
    [1, 3],
)
@pytest.mark.parametrize(
    "if_false_length",
    [1, 3],
)
def test_series_if_else_bool(
    if_true_value,
    if_false_value,
    if_true_length,
    if_false_length,
) -> None:
    if_true_series = Series.from_arrow(pa.array([if_true_value] * if_true_length, type=pa.bool_()))
    if_false_series = Series.from_arrow(pa.array([if_false_value] * if_false_length, type=pa.bool_()))
    predicate_series = Series.from_arrow(pa.array([True, False, None]))
    result = predicate_series.if_else(if_true_series, if_false_series)
    assert result.datatype() == DataType.bool()
    assert result.to_pylist() == [if_true_value, if_false_value, None]


@pytest.mark.parametrize("if_true_value", [b"Y", None])
@pytest.mark.parametrize("if_false_value", [b"N", None])
@pytest.mark.parametrize(
    "if_true_length",
    [1, 3],
)
@pytest.mark.parametrize(
    "if_false_length",
    [1, 3],
)
@pytest.mark.parametrize(
    "type, expected_type", [(pa.binary(), DataType.binary()), (pa.binary(1), DataType.fixed_size_binary(1))]
)
def test_series_if_else_binary(
    if_true_value,
    if_false_value,
    if_true_length,
    if_false_length,
    type,
    expected_type,
) -> None:
    if_true_series = Series.from_arrow(pa.array([if_true_value] * if_true_length, type=type))
    if_false_series = Series.from_arrow(pa.array([if_false_value] * if_false_length, type=type))
    predicate_series = Series.from_arrow(pa.array([True, False, None]))
    result = predicate_series.if_else(if_true_series, if_false_series)
    assert result.datatype() == expected_type
    assert result.to_pylist() == [if_true_value, if_false_value, None]


@pytest.mark.parametrize(
    ["if_true", "if_false", "expected"],
    [
        # Same length, same type
        (
            pa.array([[1, 2], [None], None, [5, 6, 7, 8]], type=pa.list_(pa.int64())),
            pa.array([[9], [10, None, 12, 13], None, [15, 16]], type=pa.list_(pa.int64())),
            [[1, 2], [10, None, 12, 13], None, [5, 6, 7, 8]],
        ),
        # Same length, different super-castable data type
        (
            pa.array([[1, 2], [None], None, [5, 6, 7, 8]], type=pa.list_(pa.int32())),
            pa.array([[9], [10, None, 12, 13], None, [15, 16]], type=pa.list_(pa.int64())),
            [[1, 2], [10, None, 12, 13], None, [5, 6, 7, 8]],
        ),
        # TODO(Clark): Uncomment this case when Arrow2 supports casting between FixedSizeList and LargeLists.
        # # Same length, different super-castable list type (FixedSizeList + List)
        # (
        #     pa.array([[1, 2], [None, 4], None, [7, 8]], type=pa.list_(pa.int64(), 2)),
        #     pa.array([[9], [10, None, 12, 13], None, [15, 16]], type=pa.list_(pa.int64())),
        #     [[1, 2], [10, None, 12, 13], None, [5, 6, 7, 8]],
        # ),
        # Broadcast left
        (
            pa.array([[1, 2]], type=pa.list_(pa.int64())),
            pa.array([[9], [10, None, 12, 13], None, [15, 16]], type=pa.list_(pa.int64())),
            [[1, 2], [10, None, 12, 13], None, [1, 2]],
        ),
        # Broadcast right
        (
            pa.array([[1, 2], [None], None, [5, 6, 7, 8]], type=pa.list_(pa.int64())),
            pa.array([[9]], type=pa.list_(pa.int64())),
            [[1, 2], [9], None, [5, 6, 7, 8]],
        ),
        # Broadcast both
        (
            pa.array([[1, 2]], type=pa.list_(pa.int64())),
            pa.array([[9]], type=pa.list_(pa.int64())),
            [[1, 2], [9], None, [1, 2]],
        ),
    ],
)
def test_series_if_else_list(if_true, if_false, expected) -> None:
    if_true_series = Series.from_arrow(if_true)
    if_false_series = Series.from_arrow(if_false)
    predicate_series = Series.from_arrow(pa.array([True, False, None, True]))
    result = predicate_series.if_else(if_true_series, if_false_series)
    assert result.datatype() == DataType.list(DataType.int64())
    assert result.to_pylist() == expected


@pytest.mark.parametrize(
    ["if_true", "if_false", "expected"],
    [
        # Same length, same type
        (
            pa.array([[1, 2], [None, 4], None, [7, 8]], type=pa.list_(pa.int64(), 2)),
            pa.array([[9, 10], [None, 12], None, [15, 16]], type=pa.list_(pa.int64(), 2)),
            [[1, 2], [None, 12], None, [7, 8]],
        ),
        # TODO(Clark): Uncomment this case when Arrow2 supports broadcasting between different FixedSizeListArrays.
        # Same length, different super-castable type
        # (pa.array([[1, 2], [None, 4], None, [7, 8]], type=pa.list_(pa.int64(), 2)), pa.array([[9, 10], [None, 12], None, [15, 16]], type=pa.list_(pa.int32(), 2))),
        # Broadcast left
        (
            pa.array([[1, 2]], type=pa.list_(pa.int64(), 2)),
            pa.array([[9, 10], [None, 12], None, [15, 16]], type=pa.list_(pa.int64(), 2)),
            [[1, 2], [None, 12], None, [1, 2]],
        ),
        # Broadcast right
        (
            pa.array([[1, 2], [None, 4], None, [7, 8]], type=pa.list_(pa.int64(), 2)),
            pa.array([[9, 10]], type=pa.list_(pa.int64(), 2)),
            [[1, 2], [9, 10], None, [7, 8]],
        ),
        # Broadcast both
        (
            pa.array([[1, 2]], type=pa.list_(pa.int64(), 2)),
            pa.array([[9, 10]], type=pa.list_(pa.int64(), 2)),
            [[1, 2], [9, 10], None, [1, 2]],
        ),
    ],
)
def test_series_if_else_fixed_size_list(if_true, if_false, expected) -> None:
    if_true_series = Series.from_arrow(if_true)
    if_false_series = Series.from_arrow(if_false)
    predicate_series = Series.from_arrow(pa.array([True, False, None, True]))
    result = predicate_series.if_else(if_true_series, if_false_series)
    assert result.datatype() == DataType.fixed_size_list(DataType.int64(), 2)
    assert result.to_pylist() == expected


@pytest.mark.parametrize(
    ["if_true", "if_false", "expected"],
    [
        # Same length, same type
        (
            pa.array(
                [[("a", 1), ("b", 2)], [("b", 3), ("c", 4)], None, [("a", 5), ("c", 7)]],
                type=pa.map_(pa.string(), pa.int64()),
            ),
            pa.array(
                [[("a", 8), ("b", 9)], [("c", 10)], None, [("a", 12), ("b", 13)]],
                type=pa.map_(pa.string(), pa.int64()),
            ),
            [[("a", 1), ("b", 2)], [("c", 10)], None, [("a", 5), ("c", 7)]],
        ),
        # Same length, different super-castable data type
        (
            pa.array(
                [[("a", 1), ("b", 2)], [("b", 3), ("c", 4)], None, [("a", 5), ("c", 7)]],
                type=pa.map_(pa.string(), pa.int64()),
            ),
            pa.array(
                [[("a", 8), ("b", 9)], [("c", 10)], None, [("a", 12), ("b", 13)]],
                type=pa.map_(pa.string(), pa.int64()),
            ),
            [[("a", 1), ("b", 2)], [("c", 10)], None, [("a", 5), ("c", 7)]],
        ),
        # Broadcast left
        (
            pa.array([[("a", 1), ("b", 2)]], type=pa.map_(pa.string(), pa.int64())),
            pa.array(
                [[("a", 8), ("b", 9)], [("c", 10)], None, [("a", 12), ("b", 13)]],
                type=pa.map_(pa.string(), pa.int64()),
            ),
            [[("a", 1), ("b", 2)], [("c", 10)], None, [("a", 1), ("b", 2)]],
        ),
        # Broadcast right
        (
            pa.array(
                [[("a", 1), ("b", 2)], [("b", 3), ("c", 4)], None, [("a", 5), ("c", 7)]],
                type=pa.map_(pa.string(), pa.int64()),
            ),
            pa.array([[("a", 8), ("b", 9)]], type=pa.map_(pa.string(), pa.int64())),
            [[("a", 1), ("b", 2)], [("a", 8), ("b", 9)], None, [("a", 5), ("c", 7)]],
        ),
        # Broadcast both
        (
            pa.array([[("a", 1), ("b", 2)]], type=pa.map_(pa.string(), pa.int64())),
            pa.array([[("a", 8), ("b", 9)]], type=pa.map_(pa.string(), pa.int64())),
            [[("a", 1), ("b", 2)], [("a", 8), ("b", 9)], None, [("a", 1), ("b", 2)]],
        ),
    ],
)
def test_series_if_else_map(if_true, if_false, expected) -> None:
    if_true_series = Series.from_arrow(if_true)
    if_false_series = Series.from_arrow(if_false)
    predicate_series = Series.from_arrow(pa.array([True, False, None, True]))
    result = predicate_series.if_else(if_true_series, if_false_series)
    assert result.datatype() == DataType.map(DataType.string(), DataType.int64())
    assert result.to_pylist() == expected


@pytest.mark.parametrize(
    ["if_true", "if_false", "expected"],
    [
        # Same length, same type
        (
            pa.array(
                [{"a": 1, "b": 2}, {"b": 3, "c": "4"}, None, {"a": 5, "b": 6, "c": "7"}],
                type=pa.struct({"a": pa.int64(), "b": pa.float64(), "c": pa.string()}),
            ),
            pa.array(
                [{"a": 8, "b": 9, "c": "10"}, {"c": "11"}, None, {"a": 12, "b": 13}],
                type=pa.struct({"a": pa.int64(), "b": pa.float64(), "c": pa.string()}),
            ),
            [{"a": 1, "b": 2.0, "c": None}, {"a": None, "b": None, "c": "11"}, None, {"a": 5, "b": 6.0, "c": "7"}],
        ),
        # TODO(Clark): Uncomment this case when Arrow2 supports casting struct types.
        # # Same length, different super-castable data type
        # (
        #     pa.array([{"a": 1, "b": 2}, {"b": 3, "c": 4}, None, {"a": 5, "b": 6, "c": 7}]).cast(pa.struct([pa.int64(), pa.float64(), pa.string()])),
        #     pa.array([{"a": 8, "b": 9, "c": 10}, {"c": 11}, None, {"a": 12, "b": 13}]).cast(pa.struct([pa.int32(), pa.float32(), pa.large_string()])),
        #     [{"a": 1, "b": 2}, {"c": 11}, None, {"a": 5, "b": 6, "c": 7}],
        # ),
        # Broadcast left
        (
            pa.array(
                [{"a": 1, "b": 2, "c": None}], type=pa.struct({"a": pa.int64(), "b": pa.float64(), "c": pa.string()})
            ),
            pa.array(
                [{"a": 8, "b": 9, "c": "10"}, {"c": "11"}, None, {"a": 12, "b": 13}],
                type=pa.struct({"a": pa.int64(), "b": pa.float64(), "c": pa.string()}),
            ),
            [{"a": 1, "b": 2.0, "c": None}, {"a": None, "b": None, "c": "11"}, None, {"a": 1, "b": 2.0, "c": None}],
        ),
        # Broadcast right
        (
            pa.array(
                [{"a": 1, "b": 2}, {"b": 3, "c": "4"}, None, {"a": 5, "b": 6, "c": "7"}],
                type=pa.struct({"a": pa.int64(), "b": pa.float64(), "c": pa.string()}),
            ),
            pa.array(
                [{"a": 8, "b": 9, "c": "10"}], type=pa.struct({"a": pa.int64(), "b": pa.float64(), "c": pa.string()})
            ),
            [{"a": 1, "b": 2.0, "c": None}, {"a": 8, "b": 9.0, "c": "10"}, None, {"a": 5, "b": 6.0, "c": "7"}],
        ),
        # Broadcast both
        (
            pa.array(
                [{"a": 1, "b": 2, "c": None}], type=pa.struct({"a": pa.int64(), "b": pa.float64(), "c": pa.string()})
            ),
            pa.array(
                [{"a": 8, "b": 9, "c": "10"}], type=pa.struct({"a": pa.int64(), "b": pa.float64(), "c": pa.string()})
            ),
            [{"a": 1, "b": 2.0, "c": None}, {"a": 8, "b": 9.0, "c": "10"}, None, {"a": 1, "b": 2.0, "c": None}],
        ),
    ],
)
def test_series_if_else_struct(if_true, if_false, expected) -> None:
    if_true_series = Series.from_arrow(if_true)
    if_false_series = Series.from_arrow(if_false)
    predicate_series = Series.from_arrow(pa.array([True, False, None, True]))
    result = predicate_series.if_else(if_true_series, if_false_series)
    assert result.datatype() == DataType.struct(
        {"a": DataType.int64(), "b": DataType.float64(), "c": DataType.string()}
    )
    assert result.to_pylist() == expected


@pytest.mark.skipif(
    get_tests_daft_runner_name() == "ray",
    reason="pyarrow extension types aren't supported on Ray clusters.",
)
@pytest.mark.parametrize(
    ["if_true_storage", "if_false_storage", "expected_storage"],
    [
        # Same length, same type
        (
            pa.array([f"{i}".encode() for i in range(4)]),
            pa.array([f"{i}".encode() for i in range(4, 8)]),
            pa.array([b"0", b"5", None, b"3"]),
        ),
        # Broadcast left
        (
            pa.array([b"0"]),
            pa.array([f"{i}".encode() for i in range(4, 8)]),
            pa.array([b"0", b"5", None, b"0"]),
        ),
        # Broadcast right
        (
            pa.array([f"{i}".encode() for i in range(4)]),
            pa.array([b"4"]),
            pa.array([b"0", b"4", None, b"3"]),
        ),
        # Broadcast both
        (
            pa.array([b"0"]),
            pa.array([b"4"]),
            pa.array([b"0", b"4", None, b"0"]),
        ),
    ],
)
def test_series_if_else_extension_type(uuid_ext_type, if_true_storage, if_false_storage, expected_storage) -> None:
    if_true_arrow = pa.ExtensionArray.from_storage(uuid_ext_type, if_true_storage)
    if_false_arrow = pa.ExtensionArray.from_storage(uuid_ext_type, if_false_storage)
    expected_arrow = pa.ExtensionArray.from_storage(uuid_ext_type, expected_storage)
    if_true_series = Series.from_arrow(if_true_arrow)
    if_false_series = Series.from_arrow(if_false_arrow)
    predicate_series = Series.from_arrow(pa.array([True, False, None, True]))

    result = predicate_series.if_else(if_true_series, if_false_series)

    assert result.datatype() == DataType.extension(
        uuid_ext_type.NAME, DataType.from_arrow_type(uuid_ext_type.storage_type), ""
    )
    result_arrow = result.to_arrow()
    assert result_arrow == expected_arrow


@pytest.mark.skipif(
    not pyarrow_supports_fixed_shape_tensor(),
    reason=f"Arrow version {ARROW_VERSION} doesn't support the canonical tensor extension type.",
)
@pytest.mark.parametrize(
    ["if_true", "if_false", "expected"],
    [
        # Same length, same type
        (
            np.arange(16).reshape((4, 2, 2)),
            np.arange(16, 32).reshape((4, 2, 2)),
            np.array([[[0, 1], [2, 3]], [[20, 21], [22, 23]], [[12, 13], [14, 15]]]),
        ),
        # Broadcast left
        (
            np.arange(4).reshape((1, 2, 2)),
            np.arange(16, 32).reshape((4, 2, 2)),
            np.array([[[0, 1], [2, 3]], [[20, 21], [22, 23]], [[0, 1], [2, 3]]]),
        ),
        # Broadcast right
        (
            np.arange(16).reshape((4, 2, 2)),
            np.arange(16, 20).reshape((1, 2, 2)),
            np.array([[[0, 1], [2, 3]], [[16, 17], [18, 19]], [[12, 13], [14, 15]]]),
        ),
        # Broadcast both
        (
            np.arange(4).reshape((1, 2, 2)),
            np.arange(16, 20).reshape((1, 2, 2)),
            np.array([[[0, 1], [2, 3]], [[16, 17], [18, 19]], [[0, 1], [2, 3]]]),
        ),
    ],
)
def test_series_if_else_canonical_tensor_extension_type(if_true, if_false, expected) -> None:
    if_true_arrow = pa.FixedShapeTensorArray.from_numpy_ndarray(if_true)
    if_false_arrow = pa.FixedShapeTensorArray.from_numpy_ndarray(if_false)
    if_true_series = Series.from_arrow(if_true_arrow)
    if_false_series = Series.from_arrow(if_false_arrow)
    predicate_series = Series.from_arrow(pa.array([True, False, None, True]))

    result = predicate_series.if_else(if_true_series, if_false_series)

    assert result.datatype() == DataType.tensor(
        DataType.from_arrow_type(if_true_arrow.type.storage_type.value_type), (2, 2)
    )
    result_arrow = result.to_arrow()

    # null element conversion to numpy is not well defined in pyarrow and changes between releases
    # so this is a workaround to ensure our tests pass regardless of the pyarrow version
    assert not result_arrow[2].is_valid
    result_array_filtered = result_arrow.filter(pa.array([True, True, False, True]))
    np.testing.assert_equal(result_array_filtered.to_numpy_ndarray(), expected)


@pytest.mark.parametrize(
    "if_true_length",
    [1, 3],
)
@pytest.mark.parametrize(
    "if_false_length",
    [1, 3],
)
def test_series_if_else_nulls(
    if_true_length,
    if_false_length,
) -> None:
    if_true_series = Series.from_arrow(pa.array([None] * if_true_length, type=pa.null()))
    if_false_series = Series.from_arrow(pa.array([None] * if_false_length, type=pa.null()))
    predicate_series = Series.from_arrow(pa.array([True, False, None]))
    result = predicate_series.if_else(if_true_series, if_false_series)
    assert result.datatype() == DataType.null()
    assert result.to_pylist() == [None, None, None]


@pytest.mark.parametrize(
    ["if_true", "if_false"],
    [
        # Same length, same type
        ([object(), object(), object()], [object(), object(), object()]),
        # Broadcast left
        ([object()], [object(), object(), object()]),
        # Broadcast right
        ([object(), object(), object()], [object()]),
        # Broadcast both
        ([object()], [object()]),
    ],
)
def test_series_if_else_python(if_true, if_false) -> None:
    if_true_series = Series.from_pylist(if_true)
    if_false_series = Series.from_pylist(if_false)
    predicate_series = Series.from_arrow(pa.array([True, False, None]))

    result = predicate_series.if_else(if_true_series, if_false_series)

    left_expected = if_true[0]
    right_expected = if_false[1] if len(if_false) > 1 else if_false[0]
    assert result.datatype() == DataType.python()
    assert result.to_pylist() == [left_expected, right_expected, None]


@pytest.mark.parametrize(
    ["predicate_value", "expected_results"], [(True, [1, 1, 1]), (False, [0, 0, 0]), (None, [None, None, None])]
)
def test_series_if_else_predicate_broadcast_numeric(predicate_value, expected_results) -> None:
    if_true_series = Series.from_arrow(pa.array([1, 1, 1], type=pa.int64()))
    if_false_series = Series.from_arrow(pa.array([0, 0, 0], type=pa.int64()))
    predicate_series = Series.from_arrow(pa.array([predicate_value], type=pa.bool_()))
    result = predicate_series.if_else(if_true_series, if_false_series)
    assert result.datatype() == DataType.int64()
    assert result.to_pylist() == expected_results


@pytest.mark.parametrize(
    ["predicate_value", "expected_results"],
    [(True, ["1", "1", "1"]), (False, ["0", "0", "0"]), (None, [None, None, None])],
)
def test_series_if_else_predicate_broadcast_strings(predicate_value, expected_results) -> None:
    if_true_series = Series.from_arrow(pa.array(["1", "1", "1"], type=pa.string()))
    if_false_series = Series.from_arrow(pa.array(["0", "0", "0"], type=pa.string()))
    predicate_series = Series.from_arrow(pa.array([predicate_value], type=pa.bool_()))
    result = predicate_series.if_else(if_true_series, if_false_series)
    assert result.datatype() == DataType.string()
    assert result.to_pylist() == expected_results


@pytest.mark.parametrize(
    ["predicate_value", "expected_results"],
    [(True, [True, True, True]), (False, [False, False, False]), (None, [None, None, None])],
)
def test_series_if_else_predicate_broadcast_bools(predicate_value, expected_results) -> None:
    if_true_series = Series.from_arrow(pa.array([True, True, True], type=pa.bool_()))
    if_false_series = Series.from_arrow(pa.array([False, False, False], type=pa.bool_()))
    predicate_series = Series.from_arrow(pa.array([predicate_value], type=pa.bool_()))
    result = predicate_series.if_else(if_true_series, if_false_series)
    assert result.datatype() == DataType.bool()
    assert result.to_pylist() == expected_results


@pytest.mark.parametrize(
    ["predicate_value", "expected_results"],
    [(True, [b"Y", b"Y", b"Y"]), (False, [b"N", b"N", b"N"]), (None, [None, None, None])],
)
@pytest.mark.parametrize(
    "type, result_type",
    [(pa.binary(), DataType.binary()), (pa.binary(1), DataType.fixed_size_binary(1))],
)
def test_series_if_else_predicate_broadcast_binary(predicate_value, expected_results, type, result_type) -> None:
    if_true_series = Series.from_arrow(pa.array([b"Y", b"Y", b"Y"], type=type))
    if_false_series = Series.from_arrow(pa.array([b"N", b"N", b"N"], type=type))
    predicate_series = Series.from_arrow(pa.array([predicate_value], type=pa.bool_()))
    result = predicate_series.if_else(if_true_series, if_false_series)
    assert result.datatype() == result_type
    assert result.to_pylist() == expected_results


@pytest.mark.parametrize(
    ["predicate_value", "expected_results"],
    [(True, [None, None, None]), (False, [None, None, None]), (None, [None, None, None])],
)
def test_series_if_else_predicate_broadcast_null(predicate_value, expected_results) -> None:
    if_true_series = Series.from_arrow(pa.array([None, None, None], type=pa.null()))
    if_false_series = Series.from_arrow(pa.array([None, None, None], type=pa.null()))
    predicate_series = Series.from_arrow(pa.array([predicate_value], type=pa.bool_()))
    result = predicate_series.if_else(if_true_series, if_false_series)
    assert result.datatype() == DataType.null()
    assert result.to_pylist() == expected_results


def test_series_if_else_wrong_types() -> None:
    if_true_series = Series.from_arrow(pa.array([1, 1, 1], type=pa.int64()))
    if_false_series = Series.from_arrow(pa.array([0, 0, 0], type=pa.int64()))
    predicate_series = Series.from_arrow(pa.array([None], type=pa.bool_()))

    with pytest.raises(ValueError):
        predicate_series.if_else(object(), if_false_series)

    with pytest.raises(ValueError):
        predicate_series.if_else(if_true_series, object())



from __future__ import annotations

import time

import pytest
import requests
from requests.exceptions import RequestException

from daft.unity_catalog import UnityCatalog


@pytest.fixture(scope="session")
def local_unity_catalog() -> UnityCatalog:
    endpoint = "http://127.0.0.1:8080"
    max_retries = 25
    retry_delay = 5  # seconds

    for attempt in range(max_retries):
        try:
            # Try to connect to the server
            response = requests.get(endpoint)
            response.raise_for_status()
            # If successful, return the UnityCatalog instance
            return UnityCatalog(endpoint=endpoint, token="not-used")
        except RequestException:
            if attempt < max_retries - 1:
                time.sleep(retry_delay)
                continue
            raise RuntimeError(f"Failed to connect to Unity Catalog server at {endpoint} after {max_retries} attempts")



from __future__ import annotations

import pytest

from daft.expressions import Expression, ExpressionsProjection, col
from daft.expressions.expressions import _resolved_col
from daft.expressions.testing import expr_structurally_equal
from daft.recordbatch import MicroPartition


def test_expressions_projection_error_dup_name():
    with pytest.raises(ValueError):
        ExpressionsProjection(
            [
                col("x"),
                col("y").alias("x"),
            ]
        )


def test_expressions_projection_empty():
    ep = ExpressionsProjection([])

    # Test len
    assert len(ep) == 0

    # Test iter
    assert list(ep) == []

    # Test eq
    assert ep == ep
    assert ep != ExpressionsProjection([col("x")])

    # Test to_name_set()
    assert ep.to_name_set() == set()

    # Test to_column_expression()
    assert ep.to_column_expressions() == ExpressionsProjection([])


def test_expressions_projection():
    exprs = [
        col("x"),
        col("y") + 1,
        col("z").alias("a"),
    ]
    ep = ExpressionsProjection(exprs)

    # Test len
    assert len(ep) == 3

    # Test iter
    for e1, e2 in zip(ep, exprs):
        assert e1.name() == e2.name()

    # Test eq
    assert ep == ep
    assert ep != ExpressionsProjection([])

    # Test to_name_set()
    assert ep.to_name_set() == {"x", "y", "a"}

    # Test to_column_expression()
    assert ep.to_column_expressions() == ExpressionsProjection([col("x"), col("y"), col("a")])


def test_expressions_union():
    exprs1 = [
        col("x"),
        col("y"),
    ]
    ep1 = ExpressionsProjection(exprs1)

    exprs2 = [
        col("z"),
    ]
    ep2 = ExpressionsProjection(exprs2)

    assert ep1.union(ep2).to_name_set() == {"x", "y", "z"}


def test_expressions_union_dup_err():
    exprs1 = [
        col("x"),
        col("y"),
    ]
    ep1 = ExpressionsProjection(exprs1)

    exprs2 = [
        col("x"),
    ]
    ep2 = ExpressionsProjection(exprs2)

    with pytest.raises(ValueError):
        assert ep1.union(ep2)


def test_expressions_union_dup_rename():
    exprs1 = [
        col("x"),
        col("y"),
    ]
    ep1 = ExpressionsProjection(exprs1)

    exprs2 = [
        col("x"),
    ]
    ep2 = ExpressionsProjection(exprs2)
    assert ep1.union(ep2, rename_dup="foo.").union(ep2, rename_dup="foo.").to_name_set() == {
        "x",
        "y",
        "foo.x",
        "foo.foo.x",
    }


def test_input_mapping():
    exprs = [
        _resolved_col("x"),
        _resolved_col("y") + 1,
        _resolved_col("z").alias("a"),
    ]
    ep = ExpressionsProjection(exprs)
    assert ep.input_mapping() == {
        "x": "x",
        "a": "z",
    }


def test_get_expression_by_name():
    exprs = [col("x")]
    ep = ExpressionsProjection(exprs)
    assert ep.get_expression_by_name("x").name() == "x"


def test_expressions_projection_indexing():
    exprs = [
        col("x"),
        col("y") + 1,
        col("z").alias("a"),
    ]
    ep = ExpressionsProjection(exprs)
    assert isinstance(ep[0], Expression)
    assert ep[0].name() == "x"
    assert isinstance(ep[:2], list)
    assert [e.name() for e in ep[:2]] == ["x", "y"]
    assert expr_structurally_equal(ep[0], col("x"))
    assert all(
        [expr_structurally_equal(result, expected) for result, expected in zip(ep[:2], [col("x"), col("y") + 1])]
    )


def test_resolve_schema():
    tbl = MicroPartition.from_pydict(
        {
            "foo": [1, 2, 3],
        }
    )
    ep = ExpressionsProjection([col("foo"), (col("foo") + 1).alias("foo_plus")])
    resolved_schema = ep.resolve_schema(tbl.schema())
    assert resolved_schema.to_name_set() == {"foo", "foo_plus"}


def test_resolve_schema_invalid_type():
    tbl = MicroPartition.from_pydict(
        {
            "foo": ["a", "b", "c"],
        }
    )
    ep = ExpressionsProjection([(col("foo") / 1).alias("invalid")])
    with pytest.raises(ValueError):
        ep.resolve_schema(tbl.schema())


def test_resolve_schema_missing_col():
    tbl = MicroPartition.from_pydict(
        {
            "foo": ["a", "b", "c"],
        }
    )
    ep = ExpressionsProjection([col("bar")])
    with pytest.raises(ValueError):
        ep.resolve_schema(tbl.schema())



from __future__ import annotations

import os

import numpy as np
import pytest

import daft
from daft import DataType, Series, col
from daft.exceptions import DaftCoreException
from tests.assets import TPCH_QUERIES


def load_tpch_queries():
    """Load all TPCH queries into a list of (name,sql) tuples."""
    queries = []
    for filename in os.listdir(TPCH_QUERIES):
        filepath = os.path.join(TPCH_QUERIES, filename)
        if os.path.isfile(filepath) and filepath.endswith(".sql"):
            with open(filepath) as f:
                sql = f.read()
                name = "TPC-H " + os.path.basename(filepath)
                queries.append((name, sql))
    return queries


def load_tpch_query(filename):
    """Load a single TPCH query from a file."""
    filepath = os.path.join(TPCH_QUERIES, filename)
    if os.path.isfile(filepath) and filepath.endswith(".sql"):
        with open(filepath) as f:
            sql = f.read()
            name = "TPC-H " + os.path.basename(filepath)
            return (name, sql)
    else:
        raise ValueError(f"File {filename} not found in {TPCH_QUERIES}")


# Load all TPCH queries once
all_tpch_queries = load_tpch_queries()


def test_sanity():
    bindings = {"test": daft.from_pydict({"a": [1, 2, 3]})}
    df = daft.sql("SELECT * FROM test", **bindings)
    assert isinstance(df, daft.DataFrame)


@pytest.mark.skip(reason="This test is a placeholder used to check that we can parse the TPC-H queries")
@pytest.mark.parametrize("name,sql", all_tpch_queries)
def test_parse_ok(name, sql):
    print(name)
    print(sql)
    print("--------------")


def test_fizzbuzz_sql():
    arr = np.arange(100)
    df = daft.from_pydict({"a": arr})
    bindings = {"test": df}
    # test case expression
    expected = daft.from_pydict(
        {
            "a": arr,
            "fizzbuzz": [
                "FizzBuzz" if x % 15 == 0 else "Fizz" if x % 3 == 0 else "Buzz" if x % 5 == 0 else str(x)
                for x in range(0, 100)
            ],
        }
    ).collect()
    df = daft.sql(
        """
    SELECT
        a,
        CASE
            WHEN a % 15 = 0 THEN 'FizzBuzz'
            WHEN a % 3 = 0 THEN 'Fizz'
            WHEN a % 5 = 0 THEN 'Buzz'
            ELSE CAST(a AS TEXT)
        END AS fizzbuzz
    FROM test
    """,
        **bindings,
    ).collect()
    assert df.to_pydict() == expected.to_pydict()


@pytest.mark.parametrize(
    "actual,expected",
    [
        ("lower(text)", daft.col("text").str.lower()),
        ("abs(n)", daft.col("n").abs()),
        ("n + 1", daft.col("n") + 1),
        ("ceil(1.1)", daft.lit(1.1).ceil()),
        ("contains(text, 'hello')", daft.col("text").str.contains("hello")),
        ("to_date(date_col, 'YYYY-MM-DD')", daft.col("date_col").str.to_date("YYYY-MM-DD")),
    ],
)
def test_sql_expr(actual, expected):
    actual = daft.sql_expr(actual)
    assert repr(actual) == repr(expected)


def test_sql_global_agg():
    df = daft.from_pydict({"n": [1, 2, 3]})
    bindings = {"test": df}
    df = daft.sql("SELECT max(n) max_n, sum(n) sum_n FROM test", **bindings)
    assert df.collect().to_pydict() == {"max_n": [3], "sum_n": [6]}
    # If there is agg and non-agg, it should fail
    # TODO: we can emit a better error message if our agg op can also include expressions on groupby expressions
    with pytest.raises(Exception):
        daft.sql("SELECT n,max(n) max_n FROM test", **bindings).collect()


@pytest.mark.parametrize(
    "query,expected",
    [
        ("SELECT sum(v) as sum FROM test GROUP BY n ORDER BY n", {"sum": [3, 7]}),
        ("SELECT n, sum(v) as sum FROM test GROUP BY n ORDER BY n", {"n": [1, 2], "sum": [3, 7]}),
        ("SELECT max(v) as max, sum(v) as sum FROM test GROUP BY n ORDER BY n", {"max": [2, 4], "sum": [3, 7]}),
        ("SELECT n as n_alias, sum(v) as sum FROM test GROUP BY n ORDER BY n", {"n_alias": [1, 2], "sum": [3, 7]}),
        ("SELECT n, sum(v) as sum FROM test GROUP BY n ORDER BY sum", {"n": [1, 2], "sum": [3, 7]}),
    ],
)
def test_sql_groupby_agg(query, expected):
    df = daft.from_pydict({"n": [1, 1, 2, 2], "v": [1, 2, 3, 4]})
    bindings = {"test": df}
    actual = daft.sql(query, **bindings)
    assert actual.collect().to_pydict() == expected


def test_sql_count_star():
    df = daft.from_pydict(
        {
            "a": ["a", "b", None, "c"],
            "b": [4, 3, 2, None],
        }
    )
    bindings = {"df": df}
    df2 = daft.sql("SELECT count(*) FROM df", **bindings)
    actual = df2.collect().to_pydict()
    expected = df.count().collect().to_pydict()
    assert actual == expected
    df2 = daft.sql("SELECT count(b) FROM df", **bindings)
    actual = df2.collect().to_pydict()
    expected = df.agg(daft.col("b").count()).collect().to_pydict()
    assert actual == expected


@pytest.fixture
def set_global_df():
    global GLOBAL_DF
    GLOBAL_DF = daft.from_pydict({"n": [1, 2, 3]})


def test_sql_function_sees_caller_tables(set_global_df):
    # sees the globals
    df = daft.sql("SELECT * FROM GLOBAL_DF")
    assert df.collect().to_pydict() == GLOBAL_DF.collect().to_pydict()
    # sees the locals
    df_copy = daft.sql("SELECT * FROM df")
    assert df.collect().to_pydict() == df_copy.collect().to_pydict()


def test_sql_function_locals_shadow_globals(set_global_df):
    GLOBAL_DF = None  # noqa: F841
    with pytest.raises(Exception, match="Table not found"):
        daft.sql("SELECT * FROM GLOBAL_DF")


def test_sql_function_globals_are_added_to_catalog(set_global_df):
    df = daft.from_pydict({"n": [1], "x": [2]})
    res = daft.sql("SELECT * FROM GLOBAL_DF g JOIN df d USING (n)", **{"df": df})
    joined = GLOBAL_DF.join(df, on="n")
    assert res.collect().to_pydict() == joined.collect().to_pydict()


def test_sql_function_catalog_is_final(set_global_df):
    df = daft.from_pydict({"a": [1]})
    # sanity check to ensure validity of below test
    assert df.collect().to_pydict() != GLOBAL_DF.collect().to_pydict()
    res = daft.sql("SELECT * FROM GLOBAL_DF", **{"GLOBAL_DF": df})
    assert res.collect().to_pydict() == df.collect().to_pydict()


def test_sql_function_register_globals(set_global_df):
    with pytest.raises(Exception, match="Table not found"):
        daft.sql("SELECT * FROM GLOBAL_DF", register_globals=False)


def test_sql_function_raises_when_cant_get_frame(monkeypatch):
    monkeypatch.setattr("inspect.currentframe", lambda: None)
    with pytest.raises(DaftCoreException, match="Cannot get caller environment"):
        daft.sql("SELECT * FROM df")


def test_sql_multi_statement_sql_error():
    with pytest.raises(Exception, match="one SQL statement allowed"):
        daft.sql("SELECT * FROM df; SELECT * FROM df")


def test_sql_tbl_alias():
    bindings = {"df": daft.from_pydict({"n": [1, 2, 3]})}
    df = daft.sql("SELECT df_alias.n FROM df AS df_alias where df_alias.n = 2", **bindings)
    assert df.collect().to_pydict() == {"n": [2]}


def test_sql_distinct():
    df = daft.from_pydict({"n": [1, 1, 2, 2]})
    df = daft.sql("SELECT DISTINCT n FROM df").collect().to_pydict()
    assert set(df["n"]) == {1, 2}


@pytest.mark.parametrize(
    "query",
    [
        "select utf8 from tbl1 order by utf8",
        "select utf8 from tbl1 order by utf8 asc",
        "select utf8 from tbl1 order by utf8 desc",
        "select utf8 as a from tbl1 order by a",
        "select utf8 as a from tbl1 order by utf8",
        "select utf8 as a from tbl1 order by utf8 asc",
        "select utf8 as a from tbl1 order by utf8 desc",
        "select utf8 from tbl1 group by utf8 order by utf8",
        "select utf8 as a from tbl1 group by utf8 order by utf8",
        "select utf8 as a from tbl1 group by a order by utf8",
        "select utf8 as a from tbl1 group by a order by a",
        "select sum(i32), utf8 as a from tbl1 group by utf8 order by a",
        "select sum(i32) as s, utf8 as a from tbl1 group by utf8 order by s",
    ],
)
def test_compiles(query):
    tbl1 = daft.from_pydict(
        {
            "utf8": ["group1", "group1", "group2", "group2"],
            "i32": [1, 2, 3, 3],
        }
    )
    bindings = {"tbl1": tbl1}
    try:
        res = daft.sql(query, **bindings)
        data = res.collect().to_pydict()
        assert data

    except Exception as e:
        print(f"Error: {e}")
        raise


def test_sql_cte():
    df = daft.from_pydict({"a": [1, 2, 3], "b": [4, 5, 6], "c": ["a", "b", "c"]})
    actual = (
        daft.sql("""
        WITH cte1 AS (select * FROM df)
        SELECT * FROM cte1
        """)
        .collect()
        .to_pydict()
    )

    expected = df.collect().to_pydict()

    assert actual == expected


def test_sql_cte_column_aliases():
    df = daft.from_pydict({"a": [1, 2, 3], "b": [4, 5, 6], "c": ["a", "b", "c"]})
    actual = (
        daft.sql("""
        WITH cte1 (cte_a, cte_b, cte_c) AS (select * FROM df)
        SELECT * FROM cte1
        """)
        .collect()
        .to_pydict()
    )

    expected = (
        df.select(
            col("a").alias("cte_a"),
            col("b").alias("cte_b"),
            col("c").alias("cte_c"),
        )
        .collect()
        .to_pydict()
    )

    assert actual == expected


def test_sql_multiple_bindings():
    df1 = daft.from_pydict({"a": [1, 2, 3], "b": [4, 5, 6], "c": ["a", "b", "c"]})
    df2 = daft.from_pydict({"x": [1, 0, 3], "y": [True, None, False], "z": [1.0, 2.0, 3.0]})
    actual = (
        daft.sql("""
        WITH
            cte1 AS (select * FROM df1),
            cte2 AS (select x as a, y, z FROM df2)
        SELECT *
        FROM cte1
        JOIN cte2 USING (a)
        """)
        .collect()
        .to_pydict()
    )
    expected = df1.join(df2.select(col("x").alias("a"), "y", "z"), on="a").collect().to_pydict()

    assert actual == expected


def test_cast_image():
    channels = 3

    data = [
        np.arange(4 * channels, dtype=np.uint8).reshape((2, 2, channels)),
        np.arange(4 * channels, 13 * channels, dtype=np.uint8).reshape((3, 3, channels)),
        None,
    ]

    s = Series.from_pylist(data, pyobj="force")
    df = daft.from_pydict({"img": s})
    actual = daft.sql("select cast(img as image(RGB)) from df", **{"df": df}).collect()
    assert actual.schema()["img"].dtype == DataType.image("RGB")



from __future__ import annotations

import os
import pathlib

import pandas as pd

import daft
from tests.conftest import assert_df_equals
from tests.cookbook.assets import COOKBOOK_DATA_CSV


def test_load(daft_df, service_requests_csv_pd_df, repartition_nparts, with_morsel_size):
    """Loading data from a CSV or Parquet works."""
    pd_slice = service_requests_csv_pd_df
    daft_slice = daft_df.repartition(repartition_nparts)
    daft_pd_df = daft_slice.to_pandas()
    assert_df_equals(daft_pd_df, pd_slice)


def test_load_csv_no_headers(tmp_path: pathlib.Path):
    """Generate a default set of headers `f0, f1, ... f{n}` when loading a CSV that has no headers."""
    csv = tmp_path / "headerless_iris.csv"
    csv.write_text("\n".join(pathlib.Path(COOKBOOK_DATA_CSV).read_text().split("\n")[1:]))
    daft_df = daft.read_csv(str(csv), has_headers=False)
    pd_df = pd.read_csv(csv, header=None, keep_default_na=False)
    pd_df.columns = [f"column_{i}" for i in range(1, 53)]
    daft_pd_df = daft_df.to_pandas()
    assert list(daft_pd_df.columns) == list(pd_df.columns)


def test_load_csv_tab_delimited(tmp_path: pathlib.Path):
    """Generate a default set of headers `col_0, col_1, ... col_{n}` when loading a CSV that has no headers."""
    csv = tmp_path / "headerless_iris.csv"
    csv.write_text(pathlib.Path(COOKBOOK_DATA_CSV).read_text().replace(",", "\t"))
    daft_df = daft.read_csv(str(csv), delimiter="\t")
    pd_df = pd.read_csv(csv, delimiter="\t")
    daft_pd_df = daft_df.to_pandas()
    assert list(daft_pd_df.columns) == list(pd_df.columns)


def test_load_json(tmp_path: pathlib.Path):
    """Generate a default set of headers `col_0, col_1, ... col_{n}` when loading a JSON file."""
    json_file = tmp_path / "iris.json"
    pd_df = pd.read_csv(COOKBOOK_DATA_CSV)
    pd_df.to_json(json_file, lines=True, orient="records")
    daft_df = daft.read_json(str(json_file))
    daft_pd_df = daft_df.to_pandas()
    assert list(daft_pd_df.columns) == list(pd_df.columns)


def test_load_pydict():
    data = {"foo": [1, 2, 3], "bar": [1.0, 2.0, 3.0], "baz": ["a", "b", "c"]}
    daft_df = daft.from_pydict(data)
    pd_df = pd.DataFrame(data)
    daft_pd_df = daft_df.to_pandas()
    assert_df_equals(daft_pd_df, pd_df, sort_key="foo")


def test_load_pylist():
    data = [
        {"foo": 1, "bar": 1.0, "baz": "a"},
        {"foo": 2, "bar": 2.0, "baz": "b"},
        {"foo": 3, "bar": 3.0, "baz": "c"},
    ]
    daft_df = daft.from_pylist(data)
    pd_df = pd.DataFrame.from_records(data)
    daft_pd_df = daft_df.to_pandas()
    assert_df_equals(daft_pd_df, pd_df, sort_key="foo")


def test_glob_files(tmpdir):
    filepaths = []
    for i in range(10):
        filepath = pathlib.Path(tmpdir) / f"file_{i}.foo"
        filepath.write_text("a" * i)
        filepaths.append(filepath)
        bar_filepath = pathlib.Path(tmpdir) / f"file_{i}.bar"
        bar_filepath.write_text("b" * i)

    daft_df = daft.from_glob_path(os.path.join(tmpdir, "*.foo"))
    daft_pd_df = daft_df.to_pandas()

    pd_df = pd.DataFrame.from_records(
        {"path": "file://" + str(path.as_posix()), "size": size, "num_rows": None}
        for path, size in zip(filepaths, list(range(10)))
    )
    pd_df = pd_df[~pd_df["path"].str.endswith(".bar")]
    pd_df = pd_df.astype({"num_rows": float})
    assert_df_equals(daft_pd_df, pd_df, sort_key="path")


def test_glob_files_single_file(tmpdir):
    filepath = pathlib.Path(tmpdir) / "file.foo"
    filepath.write_text("b" * 10)
    daft_df = daft.from_glob_path(os.path.join(tmpdir, "file.foo"))
    daft_pd_df = daft_df.to_pandas()
    pd_df = pd.DataFrame.from_records([{"path": "file://" + str(filepath), "size": 10, "num_rows": None}])
    pd_df = pd_df.astype({"num_rows": float})
    assert_df_equals(daft_pd_df, pd_df, sort_key="path")


def test_glob_files_directory(tmpdir):
    extra_empty_dir = pathlib.Path(tmpdir) / "bar"
    extra_empty_dir.mkdir()
    filepaths = []
    for i in range(10):
        for ext in ("foo", "bar"):
            filepath = pathlib.Path(tmpdir) / f"file_{i}.{ext}"
            filepath.write_text("a" * i)
            filepaths.append(filepath)

    daft_df = daft.from_glob_path(str(tmpdir))
    daft_pd_df = daft_df.to_pandas()

    listing_records = [
        {"path": "file://" + str(path.as_posix()), "size": size, "num_rows": None}
        for path, size in zip(filepaths, [i for i in range(10) for _ in range(2)])
    ]

    listing_records = listing_records
    pd_df = pd.DataFrame.from_records(listing_records)
    pd_df = pd_df.astype({"num_rows": float})
    assert_df_equals(daft_pd_df, pd_df, sort_key="path")


def test_glob_files_recursive(tmpdir):
    nested_dir_path = pathlib.Path(tmpdir) / "bar"
    nested_dir_path.mkdir()
    paths = []
    for i in range(10):
        for prefix in [pathlib.Path(tmpdir), pathlib.Path(tmpdir) / "bar"]:
            filepath = prefix / f"file_{i}.foo"
            filepath.write_text("a" * i)
            paths.append(filepath)

    daft_df = daft.from_glob_path(os.path.join(tmpdir, "**"))
    daft_pd_df = daft_df.to_pandas()
    listing_records = [
        {"path": "file://" + str(path.as_posix()), "size": size, "num_rows": None}
        for path, size in zip(paths, [i for i in range(10) for _ in range(2)])
    ]

    pd_df = pd.DataFrame.from_records(listing_records)
    pd_df = pd_df.astype({"num_rows": float})

    assert_df_equals(daft_pd_df, pd_df, sort_key="path")



from __future__ import annotations

import pyarrow as pa
import pytest

from daft.recordbatch import MicroPartition


def test_table_head() -> None:
    pa_table = pa.Table.from_pydict({"a": [1, 2, 3, 4], "b": [5, 6, 7, 8]})
    daft_recordbatch = MicroPartition.from_arrow(pa_table)
    assert len(daft_recordbatch) == 4
    assert daft_recordbatch.column_names() == ["a", "b"]

    # subslice
    headed = daft_recordbatch.head(3)
    assert len(headed) == 3
    assert headed.column_names() == ["a", "b"]
    pa_headed = headed.to_arrow()
    assert pa_table[:3] == pa_headed

    # overslice
    headed = daft_recordbatch.head(5)
    assert len(headed) == 4
    assert headed.column_names() == ["a", "b"]
    pa_headed = headed.to_arrow()
    assert pa_table == pa_headed

    # negative slice
    with pytest.raises(ValueError, match="negative number"):
        headed = daft_recordbatch.head(-1)



from __future__ import annotations

from typing import TYPE_CHECKING

from daft.daft import eq as _eq

if TYPE_CHECKING:
    from daft.expressions import Expression


def expr_structurally_equal(e1: Expression, e2: Expression) -> bool:
    """Returns a boolean indicating whether two Expressions are structurally equal.

    Structurally equal is defined as:
    1. Expressions' local parameters are value-wise equal
    2. Expressions have the same number of children Expressions
    3. (Recursive) Expressions' children are structurally equal to each other as well.
    """
    return _eq(e1._expr, e2._expr)
